"""
Some JSON encoding and decoding utilities.
"""

from __future__ import absolute_import, division, print_function

import json
from datetime import date,datetime,time
import base64
from typing import Any
import zlib
import logging
import ast
import inspect

from iceprod.core import dataclasses
from iceprod.core import util

logger = logging.getLogger('jsonUtil')


class json_compressor:
    """Used for files and other large things sent over json.
       Great for log files.
    """
    @staticmethod
    def compress(obj):
        return base64.b64encode(zlib.compress(obj)) if obj else b''

    @staticmethod
    def uncompress(obj):
        return zlib.decompress(base64.b64decode(obj)).decode('utf-8') if obj else ''


class datetime_converter:
    @staticmethod
    def dumps(obj):
        return obj.isoformat()

    @staticmethod
    def loads(obj,name=None):
        if ':' in obj:
            if 'T' in obj or ' ' in obj:
                center = ' '
                if 'T' in obj:
                    center = 'T'
                # must be datetime
                if '.' in obj:
                    return datetime.strptime(obj, "%Y-%m-%d"+center+"%H:%M:%S.%f")
                else:
                    return datetime.strptime(obj, "%Y-%m-%d"+center+"%H:%M:%S")
            else:
                # must be time
                if '.' in obj:
                    return datetime.strptime(obj, "%H:%M:%S.%f")
                else:
                    return datetime.strptime(obj, "%H:%M:%S")
        else:
            # must be date
            return datetime.strptime(obj, "%Y-%m-%d")


class date_converter(datetime_converter):
    @staticmethod
    def loads(obj,name=None):
        d = datetime_converter.loads(obj)
        return date(d.year,d.month,d.day)


class time_converter(datetime_converter):
    @staticmethod
    def loads(obj,name=None):
        d = datetime_converter.loads(obj)
        return time(d.hour,d.minute,d.second,d.microsecond)


class binary_converter:
    """note that is is really only for decode of json, since python bytes are strings"""
    @staticmethod
    def dumps(obj,name=None):
        return base64.b64encode(obj)

    @staticmethod
    def loads(obj,name=None):
        return base64.b64decode(obj).decode('utf-8')


class bytearray_converter:
    @staticmethod
    def dumps(obj,name=None):
        return base64.b64encode(str(obj))

    @staticmethod
    def loads(obj,name=None):
        return bytearray(base64.b64decode(obj))


class set_converter:
    @staticmethod
    def dumps(obj):
        return list(obj)

    @staticmethod
    def loads(obj,name=None):
        return set(obj)


# do some dataclass json conversions


class var_converter:
    @staticmethod
    def dumps(obj):
        return obj.__dict__

    @staticmethod
    def loads(obj,name=None):
        ret = getattr(dataclasses,name)()
        for k in obj:
            setattr(ret,k,obj[k])
        return ret


# convert the IFace


class iface_converter:
    @staticmethod
    def dumps(obj):
        return obj.__dict__

    @staticmethod
    def loads(obj,name=None):
        ret = util.IFace()
        for k in obj:
            setattr(ret,k,obj[k])
        return ret


# do some default conversions
# for things like OrderedDict


class repr_converter:
    @staticmethod
    def dumps(obj):
        return repr(obj)

    @staticmethod
    def loads(obj,name=None):
        parts = obj.split('(',1)
        type = parts[0]
        if type not in globals():
            raise Exception()
        parts2 = parts[1].rsplit(')',1)
        args = ast.literal_eval(parts2[0])
        if isinstance(args,tuple):
            ret = globals()['type'](*args)
        else:
            ret = globals()['type'](args)
        return ret


JSONConverters = {
    'datetime':datetime_converter,
    'date':date_converter,
    'time':time_converter,
    'binary':binary_converter,
    'bytearray':bytearray_converter,
    'OrderedDict':repr_converter,
    'set':set_converter,
    'IFace':iface_converter,
}
for k in dict(inspect.getmembers(dataclasses,inspect.isclass)):
    JSONConverters[k] = var_converter


def objToJSON(obj):
    if isinstance(obj,(dict,list,tuple,str,int,float,bool)) or obj is None:
        return obj
    else:
        name = obj.__class__.__name__
        if name in JSONConverters:
            return {'__jsonclass__':[name,JSONConverters[name].dumps(obj)]}
        else:
            logger.error('name: %s, obj: %r', name, obj)
            raise Exception('Cannot encode %s class to JSON'%name)


def JSONToObj(obj):
    ret = obj
    if isinstance(obj,dict) and '__jsonclass__' in obj:
        logger.info('try unpacking class')
        try:
            name = obj['__jsonclass__'][0]
            if name not in JSONConverters:
                raise Exception('class %r not found in converters'%name)
            obj_repr = obj['__jsonclass__'][1]
            ret = JSONConverters[name].loads(obj_repr,name=name)
        except Exception as e:
            logger.warning('error making json class: %r',e,exc_info=True)
    return ret


# copied from tornado.escape so we don't have to include that project
def recursive_unicode(obj):
    """Walks a simple data structure, converting byte strings to unicode.

    Supports lists, tuples, sets, and dictionaries.
    """
    if isinstance(obj, dict):
        return {recursive_unicode(k): recursive_unicode(obj[k]) for k in obj}
    elif isinstance(obj, set):
        return {recursive_unicode(i) for i in obj}
    elif isinstance(obj, list):
        return [recursive_unicode(i) for i in obj]
    elif isinstance(obj, tuple):
        return tuple(recursive_unicode(i) for i in obj)
    elif isinstance(obj, bytes):
        return obj.decode("utf-8")
    else:
        return obj


def json_encode(value: dict[Any, Any] | list[Any], indent=None):
    """JSON-encodes the given Python object."""
    return json.dumps(recursive_unicode(value),default=objToJSON,separators=(',',':'), indent=indent).replace("</", "<\\/")


def json_decode(value: str):
    """Returns Python objects for the given JSON string."""
    return json.loads(value,object_hook=JSONToObj)



from functools import partial
from iceprod.server import states


def test_get_all_prev_statuses():
    st = {
        'a': ['b'],
        'b': ['c'],
        'c': [],
    }
    assert states.get_all_prev_statuses(st, 'b') == ['a']


def test_dataset_prev_status():
    assert states.dataset_prev_statuses('complete') == ['processing']
    assert states.dataset_prev_statuses.cache_info().hits == 0
    states.dataset_prev_statuses('complete')
    assert states.dataset_prev_statuses.cache_info().hits == 1


def test_status_sort():
    st = {
        'a': ['b'],
        'b': ['c'],
        'c': [],
    }
    assert sorted(['c', 'a'], key=partial(states.status_sort, st)) == ['a', 'c']



"""
Test script for core exe
"""

import logging
import pytest

import iceprod.core.config
import iceprod.core.exe
from iceprod.core.exe import Data, Transfer
from iceprod.core.defaults import add_default_options


logger = logging.getLogger('exe_test')


def get_task(config):
    d = iceprod.core.config.Dataset('did123', 123, 2, 1, 1, 'processing', 0.5, 'grp', 'usr', False, config)
    d.fill_defaults()
    add_default_options(d.config['options'])
    j = iceprod.core.config.Job(d, 'j123', 1, 'processing')
    t = iceprod.core.config.Task(d, j, 't123', 0, 'foo', [], {}, 'waiting', '', {})
    return t


def test_config_parser():
    t = get_task({
        'steering': {
            'parameters': {'foo': 1, 'bar': [2, 3, 4]}
        },
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{}]
            }]
        }]
    })

    c = iceprod.core.exe.ConfigParser(t.dataset, logger=logger)
    assert c.parseValue('$steering(foo)') == 1
    assert c.parseObject({'foo': '$steering(bar)'}, {}) == {'foo': [2, 3, 4]}
    assert c.parseObject('$(bar)', {'parameters': {'bar': {'a': 'b'}}}) == {'a': 'b'}


def test_scope_env():
    t = get_task({
        'steering': {
            'parameters': {'foo': 1, 'bar': [2, 3, 4]}
        },
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{}]
            }],
            'data': [{
                'movement': 'input',
                'remote': 'https://foo.bar/baz'
            }],
        }]
    })

    c = iceprod.core.exe.ConfigParser(t.dataset, logger=logger)
    with iceprod.core.exe.scope_env(c, t.dataset.config['steering']) as env:
        # test parameters
        assert env.parameters == t.dataset.config['steering']['parameters']
        assert c.parseObject('$(foo)', env.to_parser_dict()) == 1

        # test parsing data
        with iceprod.core.exe.scope_env(c, t.dataset.config['tasks'][0], env) as tenv:
            assert tenv.input_files == [Data('https://foo.bar/baz', 'baz', Transfer.TRUE)]
        assert env.input_files == []


def test_download_data():
    data = {
        'movement': 'input',
        'remote': 'https://foo.bar/baz'
    }
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{}]
            }],
            'data': [data],
        }]
    })

    c = iceprod.core.exe.ConfigParser(t.dataset, logger=logger)
    ret = iceprod.core.exe.downloadData(data, c, logger=logger)
    assert ret == Data('https://foo.bar/baz', 'baz', Transfer.TRUE)


def test_download_data_false():
    data = {
        'movement': 'input',
        'remote': 'https://foo.bar/baz',
        'transfer': 0,
    }
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{}]
            }],
            'data': [data],
        }]
    })

    c = iceprod.core.exe.ConfigParser(t.dataset, logger=logger)
    ret = iceprod.core.exe.downloadData(data, c, logger=logger)
    assert not ret


def test_download_data_maybe():
    data = {
        'movement': 'input',
        'remote': 'https://foo.bar/baz',
        'transfer': 'maybe',
    }
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{}]
            }],
            'data': [data],
        }]
    })

    c = iceprod.core.exe.ConfigParser(t.dataset, logger=logger)
    ret = iceprod.core.exe.downloadData(data, c, logger=logger)
    assert ret == Data('https://foo.bar/baz', 'baz', Transfer.MAYBE)


def test_download_data_invalid():
    data = {
        'movement': 'input',
        'remote': ''
    }
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{}]
            }],
            'data': [data],
        }]
    })

    c = iceprod.core.exe.ConfigParser(t.dataset, logger=logger)
    with pytest.raises(Exception):
        iceprod.core.exe.downloadData(data, c, logger=logger)


def test_download_data_no_transfer():
    data = {
        'movement': 'input',
        'remote': 'https://foo.bar/baz',
        'transfer': 'no'
    }
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{}]
            }],
            'data': [data],
        }]
    })

    c = iceprod.core.exe.ConfigParser(t.dataset, logger=logger)
    ret = iceprod.core.exe.downloadData(data, c, logger=logger)
    assert not ret


def test_download_data_job_temp():
    data = {
        'movement': 'input',
        'local': 'baz',
        'remote': '',
        'type': 'job_temp'
    }
    t = get_task({
        'options': {
            'job_temp': 'https://foo.bar',
        },
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{}]
            }],
            'data': [data],
        }]
    })

    c = iceprod.core.exe.ConfigParser(t.dataset, logger=logger)
    ret = iceprod.core.exe.downloadData(data, c, logger=logger)
    assert ret == Data('https://foo.bar/baz', 'baz', Transfer.TRUE)


def test_upload_data():
    data = {
        'movement': 'output',
        'remote': 'https://foo.bar/baz'
    }
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{}]
            }],
            'data': [data],
        }]
    })

    c = iceprod.core.exe.ConfigParser(t.dataset, logger=logger)
    ret = iceprod.core.exe.uploadData(data, c, logger=logger)
    assert ret == Data('https://foo.bar/baz', 'baz', Transfer.TRUE)


def test_upload_data_invalid():
    data = {
        'movement': 'output',
        'remote': ''
    }
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{}]
            }],
            'data': [data],
        }]
    })

    c = iceprod.core.exe.ConfigParser(t.dataset, logger=logger)
    with pytest.raises(Exception):
        iceprod.core.exe.uploadData(data, c, logger=logger)


def test_upload_data_no_transfer():
    data = {
        'movement': 'output',
        'remote': 'https://foo.bar/baz',
        'transfer': 'no'
    }
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{}]
            }],
            'data': [data],
        }]
    })

    c = iceprod.core.exe.ConfigParser(t.dataset, logger=logger)
    ret = iceprod.core.exe.uploadData(data, c, logger=logger)
    assert not ret


def test_upload_data_job_temp():
    data = {
        'movement': 'output',
        'local': 'baz',
        'type': 'job_temp'
    }
    t = get_task({
        'options': {
            'job_temp': 'https://foo.bar',
        },
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{}]
            }],
            'data': [data],
        }]
    })

    c = iceprod.core.exe.ConfigParser(t.dataset, logger=logger)
    ret = iceprod.core.exe.uploadData(data, c, logger=logger)
    assert ret == Data('https://foo.bar/baz', 'baz', Transfer.TRUE)


async def test_write_to_script_no_module(tmp_path):
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{}]
            }],
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    with pytest.raises(iceprod.core.exe.ConfigError):
        await ws.convert()
    assert not ws.infiles
    assert not ws.outfiles


async def test_write_to_script_module_src(tmp_path):
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': False,
                    'src': 'foo.py'
                }]
            }],
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    scriptpath = await ws.convert()

    assert not ws.infiles
    assert not ws.outfiles
    script = open(scriptpath).read()
    lines = [line for line in script.split('\n') if not (not line.strip() or line.startswith('#') or line.startswith('set '))]
    assert lines[-1] == 'python foo.py'


async def test_write_to_script_module_shell(tmp_path):
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': False,
                    'src': 'foo.sh'
                }]
            }],
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    scriptpath = await ws.convert()

    assert not ws.infiles
    assert not ws.outfiles
    script = open(scriptpath).read()
    lines = [line for line in script.split('\n') if not (not line.strip() or line.startswith('#') or line.startswith('set '))]
    assert lines[-1] == '/bin/sh foo.sh'


async def test_write_to_script_module_binary(tmp_path):
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': False,
                    'src': 'foo'
                }]
            }],
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    scriptpath = await ws.convert()

    assert not ws.infiles
    assert not ws.outfiles
    script = open(scriptpath).read()
    lines = [line for line in script.split('\n') if not (not line.strip() or line.startswith('#') or line.startswith('set '))]
    assert lines[-1] == './foo'

async def test_write_to_script_module_binary_fullpath(tmp_path):
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': False,
                    'src': '/cvmfs/foo'
                }]
            }],
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    scriptpath = await ws.convert()

    assert not ws.infiles
    assert not ws.outfiles
    script = open(scriptpath).read()
    lines = [line for line in script.split('\n') if not (not line.strip() or line.startswith('#') or line.startswith('set '))]
    assert lines[-1] == '/cvmfs/foo'


async def test_write_to_script_os_arch(tmp_path):
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': False,
                    'src': '/cvmfs/foo/$environ(OS_ARCH)'
                }]
            }],
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    scriptpath = await ws.convert()

    assert not ws.infiles
    assert not ws.outfiles
    script = open(scriptpath).read()
    lines = [line for line in script.split('\n') if not (not line.strip() or line.startswith('#') or line.startswith('set '))]
    assert lines[-1] == '/cvmfs/foo/$OS_ARCH'


async def test_write_to_script_tray_iter(tmp_path):
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'iterations': 3,
                'modules': [{
                    'env_clear': False,
                    'src': 'foo.py',
                    'args': '$(iter)'
                }]
            }],
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    scriptpath = await ws.convert()

    assert not ws.infiles
    assert not ws.outfiles
    script = open(scriptpath).read()
    logging.debug('script: \n%s', script)
    lines = [line for line in script.split('\n') if not (not line.strip() or line.startswith('#') or line.startswith('set '))]
    assert lines[-3:] == [
        'python foo.py 0',
        'python foo.py 1',
        'python foo.py 2',
    ]


async def test_write_to_script_module_env_clear(tmp_path):
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': True,
                    'src': 'foo.py'
                }]
            }],
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    scriptpath = await ws.convert()

    assert not ws.infiles
    assert not ws.outfiles
    script = open(scriptpath).read()
    lines = [line for line in script.split('\n') if not (not line.strip() or line.startswith('#') or line.startswith('set '))]
    assert lines[-1].startswith('env -i ')
    assert lines[-1].endswith(' python foo.py')


async def test_write_to_script_module_env_shell(tmp_path):
    t = get_task({
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': False,
                    'env_shell': '/foo/bar/baz.sh',
                    'src': 'foo.py'
                }]
            }],
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    scriptpath = await ws.convert()

    assert not ws.infiles
    assert not ws.outfiles
    script = open(scriptpath).read()
    lines = [line for line in script.split('\n') if not (not line.strip() or line.startswith('#') or line.startswith('set '))]
    assert lines[-1] == '/foo/bar/baz.sh python foo.py'


async def test_write_to_script_data(tmp_path):
    t = get_task({
        'options': {
            'job_temp': 'https://foo.bar',
        },
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': False,
                    'src': 'foo.py'
                }]
            }],
            'data': [{
                'movement': 'input',
                'type': 'permanent',
                'remote': 'https://foo.bar/baz',
            }, {
                'movement': 'output',
                'type': 'job_temp',
                'local': '1234',
            }]
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    scriptpath = await ws.convert()

    assert ws.infiles == {Data('https://foo.bar/baz', 'baz', Transfer.TRUE)}
    assert ws.outfiles == {Data('https://foo.bar/1234', '1234', Transfer.TRUE)}
    script = open(scriptpath).read()
    lines = [line for line in script.split('\n') if not (not line.strip() or line.startswith('#') or line.startswith('set '))]
    assert lines[-1] == 'python foo.py'


async def test_write_to_script_data_task_files(tmp_path):
    t = get_task({
        'options': {
            'job_temp': 'https://foo.bar',
        },
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': False,
                    'src': 'foo.py',
                    'args': '--foo -b $(input) $(output)',
                }]
            }],
            'data': [{
                'movement': 'input',
                'type': 'permanent',
                'remote': 'https://foo.bar/baz',
            }, {
                'movement': 'output',
                'type': 'job_temp',
                'local': '1234',
            }]
        }]
    })
    t.task_files = [{
        'movement': 'input',
        'type': 'permanent',
        'remote': 'https://foo.bar/blah',
        'local': '',
        'transfer': True,
    }, {
        'movement': 'output',
        'type': 'job_temp',
        'remote': '',
        'local': 'abcde',
        'transfer': True,
    }]

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    scriptpath = await ws.convert()

    assert ws.infiles == {Data('https://foo.bar/baz', 'baz', Transfer.TRUE), Data('https://foo.bar/blah', 'blah', Transfer.TRUE)}
    assert ws.outfiles == {Data('https://foo.bar/1234', '1234', Transfer.TRUE), Data('https://foo.bar/abcde', 'abcde', Transfer.TRUE)}
    script = open(scriptpath).read()
    lines = [line for line in script.split('\n') if not (not line.strip() or line.startswith('#') or line.startswith('set '))]
    assert lines[-1] == 'python foo.py --foo -b baz blah 1234 abcde'


async def test_write_to_script_data_dups(tmp_path):
    t = get_task({
        'options': {
            'job_temp': 'https://foo.bar',
        },
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': False,
                    'src': 'foo.py',
                    'data': [{
                        'movement': 'input',
                        'type': 'permanent',
                        'remote': 'https://foo.bar/baz',
                    }, {
                        'movement': 'output',
                        'type': 'job_temp',
                        'local': '1234',
                    }]
                }],
                'data': [{
                    'movement': 'input',
                    'type': 'permanent',
                    'remote': 'https://foo.bar/baz',
                }]
            }],
            'data': [{
                'movement': 'input',
                'type': 'permanent',
                'remote': 'https://foo.bar/baz',
            }]
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    scriptpath = await ws.convert()

    assert ws.infiles == {Data('https://foo.bar/baz', 'baz', Transfer.TRUE)}
    assert ws.outfiles == {Data('https://foo.bar/1234', '1234', Transfer.TRUE)}


async def test_write_to_script_data_iterations(tmp_path):
    t = get_task({
        'options': {
            'job_temp': 'https://foo.bar',
        },
        'steering': {
            'parameters': {
                'foo0': 'baz',
                'test': 'foo_bar_$steering(foo$(iter))',
                'remote': 'https://foo.bar',
            }
        },
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': False,
                    'src': 'foo.py'
                }],
                'data': [{
                    'movement': 'input',
                    'type': 'permanent',
                    'remote': '$steering(remote)/$steering(test)',
                }]
            }]
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    scriptpath = await ws.convert()

    assert ws.infiles == {Data('https://foo.bar/foo_bar_baz', 'foo_bar_baz', Transfer.TRUE)}
    assert ws.outfiles == set()


async def test_write_to_script_data_transfer(tmp_path):
    t = get_task({
        'options': {
            'job_temp': 'https://foo.bar',
        },
        'steering': {
            'parameters': {
                'foo0': 'baz',
                'test': 'foo_bar_$steering(foo$(iter))',
                'remote': 'https://foo.bar',
            }
        },
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': False,
                    'src': 'foo.py',
                    'data': [{
                        'movement': 'input',
                        'type': 'permanent',
                        'remote': 'gsiftp://foo.bar/baz',
                        'local': 'foo_local',
                    }, {
                        'movement': 'output',
                        'type': 'permanent',
                        'transfer': 'maybe',
                        'remote': 'gsiftp://foo.bar/stats',
                        'local': 'stats',
                    }, {
                        'movement': 'output',
                        'type': 'permanent',
                        'transfer': True,
                        'remote': 'gsiftp://foo.bar/stats2',
                        'local': 'stats2',
                    }, {
                        'movement': 'both',
                        'type': 'permanent',
                        'transfer': 'maybe',
                        'remote': 'osdf://icecube/wipac/data/file'
                    }]
                }],
                'data': [{
                    'movement': 'input',
                    'type': 'permanent',
                    'remote': '$steering(remote)/$steering(test)',
                }]
            }]
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    scriptpath = await ws.convert(transfer=True)

    script = open(scriptpath).read()
    assert 'gsiftp://foo.bar/baz' in script
    assert 'foo_local' in script
    assert 'gsiftp://foo.bar/stats' in script
    assert 'if [ -f stats ' in script
    assert 'gsiftp://foo.bar/stats2' in script
    assert 'if [ -f stats2 ' not in script
    assert 'object get' in script
    assert 'object put' in script
    assert ' osdf://icecube/wipac/data/file ' in script
    assert 'if [ -f file ' in script

    assert ws.infiles == {Data('https://foo.bar/foo_bar_baz', 'foo_bar_baz', Transfer.TRUE)}
    assert ws.outfiles == set()


async def test_write_to_script_executable(tmp_path):
    t = get_task({
        'options': {
            'job_temp': 'https://foo.bar',
        },
        'steering': {
            'parameters': {
                'foo0': 'baz',
                'test': 'foo_bar_$steering(foo$(iter))',
                'remote': 'https://foo.bar',
            }
        },
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': False,
                    'src': '$steering(remote)/foo.py',
                }]
            }]
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    await ws.convert(transfer=True)

    assert ws.infiles == {Data('https://foo.bar/foo.py', 'foo.py', Transfer.TRUE)}
    assert ws.outfiles == set()


async def test_write_to_script_configs(tmp_path):
    t = get_task({
        'options': {
            'job_temp': 'https://foo.bar',
        },
        'steering': {
            'parameters': {
                'foo0': 'baz',
                'test': 'foo_bar_$steering(foo$(iter))',
                'remote': 'https://foo.bar',
            }
        },
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': False,
                    'src': 'foo.py',
                    'configs': {
                        'snowstorm.json': {
                            'foo': 1,
                            'bar': 2,
                            'baz': {'remote': '$steering(remote)'},
                        },
                    },
                }]
            }]
        }]
    })

    ws = iceprod.core.exe.WriteToScript(t, workdir=tmp_path, logger=logger)
    scriptpath = await ws.convert(transfer=True)

    config_path = tmp_path / 'snowstorm.json'
    assert config_path.exists()
    assert ws.infiles == {Data(str(config_path), 'snowstorm.json', Transfer.TRUE)}
    assert ws.outfiles == set()



import pytest
import requests


async def test_server_bad_route(server):
    client = server(roles=['system'])
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('GET', '/foo')
    assert exc_info.value.response.status_code == 404

async def test_server_bad_method(server):
    client = server(roles=['system'])
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', '/healthz')
    assert exc_info.value.response.status_code == 405

async def test_server_health(server):
    client = server()
    ret = await client.request('GET', '/healthz')
    assert ret['num_requests'] == 0



"""
Test script for dataset_prio
"""

import logging
import uuid

from iceprod.server import dataset_prio

logger = logging.getLogger('dataset_prio_test')


def test_20_calc_dataset_prio():
    """Test calc_dataset_prio"""
    dataset = {'dataset_id': uuid.uuid1().hex,
               'dataset': 1,
               'tasks_submitted': 0,
               'priority': 0,
              }
    prio1 = dataset_prio.calc_dataset_prio(dataset)
    if not isinstance(prio1,(int,float)):
        raise Exception('dataset prio is not a number')

    dataset = {'dataset_id': uuid.uuid1().hex,
               'dataset': 1,
               'tasks_submitted': 0,
               'priority': 1,
              }
    prio2 = dataset_prio.calc_dataset_prio(dataset)
    if prio2 < prio1:
        raise Exception('priority is not winning')

    dataset = {'dataset_id': uuid.uuid1().hex,
               'dataset': 1,
               'tasks_submitted': 100,
               'priority': 1,
              }
    prio3 = dataset_prio.calc_dataset_prio(dataset)
    if prio2 < prio3:
        raise Exception('greater # tasks submitted is not losing')

    dataset = {'dataset_id': uuid.uuid1().hex,
               'dataset': 2,
               'tasks_submitted': 0,
               'priority': 1,
              }
    prio4 = dataset_prio.calc_dataset_prio(dataset)
    if prio2 < prio4:
        raise Exception('greater dataset_id is not losing')

    dataset = {'dataset_id': uuid.uuid1().hex,
               'dataset': 1,
               'tasks_submitted': 0,
               'priority': -1,
              }
    prio5 = dataset_prio.calc_dataset_prio(dataset)
    if prio5 != prio1:
        raise Exception('negative prio not reset to 0')


def test_21_calc_datasets_prios():
    """Test calc_datasets_prios"""
    dataset = {'dataset_id': uuid.uuid1().hex,
               'dataset': 1,
               'tasks_submitted': 0,
               'priority': 0,
              }
    dataset2 = {'dataset_id': uuid.uuid1().hex,
               'dataset': 2,
               'tasks_submitted': 0,
               'priority': 0,
              }
    datasets = {dataset['dataset_id']:dataset,
               dataset2['dataset_id']:dataset2}

    prios = dataset_prio.calc_datasets_prios(datasets)
    for p in prios.values():
        if not isinstance(p,(int,float)):
            raise Exception('dataset prio is not a number')
    if prios[dataset['dataset_id']] != prios[dataset2['dataset_id']]:
        logger.info(prios)
        raise Exception('datasets not equal in priority')

    dataset = {'dataset_id': uuid.uuid1().hex,
               'dataset': 1,
               'tasks_submitted': 0,
               'priority': 1,
              }
    dataset2 = {'dataset_id': uuid.uuid1().hex,
               'dataset': 2,
               'tasks_submitted': 0,
               'priority': 1,
              }
    datasets = {dataset['dataset_id']:dataset,
               dataset2['dataset_id']:dataset2}

    prios = dataset_prio.calc_datasets_prios(datasets)
    for p in prios.values():
        if not isinstance(p,(int,float)):
            raise Exception('dataset prio is not a number')
    if prios[dataset['dataset_id']] <= prios[dataset2['dataset_id']]:
        logger.info(prios)
        raise Exception('datasets in wrong order')




import random
import string

import pytest



async def test_rest_logs_post(server):
    client = server(roles=['system'])

    data = {'data':'foo bar baz'}
    ret = await client.request('POST', '/logs', data)
    log_id = ret['result']


async def test_rest_logs_get(server):
    client = server(roles=['system'])

    data = {'name': 'stdlog', 'data': 'foo bar baz'}
    ret = await client.request('POST', '/logs', data)
    log_id = ret['result']

    ret = await client.request('GET', '/logs')
    assert log_id in ret
    assert len(ret) == 1
    for k in data:
        assert k in ret[log_id]
        assert data[k] == ret[log_id][k]

    args = {'name': 'stdlog', 'keys': 'log_id|name|data'}
    ret = await client.request('GET', '/logs', args)
    assert log_id in ret


async def test_rest_logs_get_details(server):
    client = server(roles=['system'])

    data = {'data': 'foo bar baz'}
    ret = await client.request('POST', '/logs', data)
    log_id = ret['result']

    ret = await client.request('GET', f'/logs/{log_id}')
    for k in data:
        assert k in ret
        assert data[k] == ret[k]


async def test_rest_logs_dataset_post(server):
    client = server(roles=['user'], groups=['users'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    data = {'data':'foo bar baz'}
    ret = await client.request('POST', f'/datasets/{dataset_id}/logs', data)
    log_id = ret['result']


async def test_rest_logs_dataset_get(server):
    client = server(roles=['user'], groups=['users'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    data = {'data':'foo bar baz', 'dataset_id': dataset_id}
    ret = await client.request('POST', f'/datasets/{dataset_id}/logs', data)
    log_id = ret['result']

    ret = await client.request('GET', f'/datasets/{dataset_id}/logs/{log_id}')
    for k in data:
        assert k in ret
        assert data[k] == ret[k]


async def test_rest_logs_dataset_task_get(server):
    client = server(roles=['user'], groups=['users'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    data = {'data':'foo', 'dataset_id': dataset_id, 'task_id': 'bar', 'name': 'stdout'}
    ret = await client.request('POST', f'/datasets/{dataset_id}/logs', data)
    log_id = ret['result']

    ret = await client.request('GET', f'/datasets/{dataset_id}/tasks/bar/logs')
    assert 'logs' in ret
    assert len(ret['logs']) == 1
    assert ret['logs'][0]['log_id'] == log_id
    assert data['data'] == ret['logs'][0]['data']

    # now try for groupings (only last of named type)
    data = {'data':'bar', 'dataset_id': dataset_id, 'task_id': 'bar', 'name': 'stderr'}
    await client.request('POST', f'/datasets/{dataset_id}/logs', data)
    
    data = {'data':'baz', 'dataset_id': dataset_id, 'task_id': 'bar', 'name': 'stdout'}
    await client.request('POST', f'/datasets/{dataset_id}/logs', data)
    
    ret = await client.request('GET', f'/datasets/{dataset_id}/tasks/bar/logs', {'group': 'true'})
    assert 'logs' in ret
    assert len(ret['logs']) == 2
    assert ret['logs'][0]['data'] == 'baz'
    assert ret['logs'][1]['data'] == 'bar'

    # now check order, num, and keys
    ret = await client.request('GET', f'/datasets/{dataset_id}/tasks/bar/logs', {'order': 'asc', 'num': 1, 'keys': 'log_id|data'})
    assert 'logs' in ret
    assert len(ret['logs']) == 1
    assert ret['logs'][0]['log_id'] == log_id
    assert ret['logs'][0]['data'] == 'foo'


def fake_data(N):
    return ''.join(random.choices(string.printable, k=N))


async def test_rest_logs_s3_post(s3conn, server):
    client = server(roles=['system'])

    data = {'data': fake_data(2000000)}
    ret = await client.request('POST', '/logs', data)
    log_id = ret['result']

    body = s3conn.get_object(Bucket='iceprod2-logs', Key=log_id)['Body'].read().decode('utf-8')
    assert body == data['data']

    data = {'data': fake_data(200000)}
    ret = await client.request('POST', '/logs', data)
    log_id = ret['result']

    with pytest.raises(Exception):
        s3conn.get_object(Bucket='iceprod2-logs', Key=log_id)


async def test_rest_logs_s3_get(s3conn, server):
    client = server(roles=['system'])

    data = {'data': fake_data(2000000)}
    ret = await client.request('POST', '/logs', data)
    log_id = ret['result']

    body = s3conn.get_object(Bucket='iceprod2-logs', Key=log_id)['Body'].read().decode('utf-8')
    assert body == data['data']

    ret = await client.request('GET', f'/logs/{log_id}')
    assert ret['data'] == data['data']

    data = {'data': fake_data(200000)}
    ret = await client.request('POST', '/logs', data)
    log_id = ret['result']

    with pytest.raises(Exception):
        s3conn.get_object(Bucket='iceprod2-logs', Key=log_id)

    ret = await client.request('GET', f'/logs/{log_id}')
    assert ret['data'] == data['data']


async def test_rest_logs_s3_dataset_post(s3conn, server):
    client = server(roles=['user'], groups=['users'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    data = {'data': fake_data(2000000)}
    ret = await client.request('POST', f'/datasets/{dataset_id}/logs', data)
    log_id = ret['result']

    body = s3conn.get_object(Bucket='iceprod2-logs', Key=log_id)['Body'].read().decode('utf-8')
    assert body == data['data']

    ret = await client.request('GET', f'/datasets/{dataset_id}/logs/{log_id}')
    assert ret['data'] == data['data']

    data = {'data': fake_data(200000)}
    ret = await client.request('POST', f'/datasets/{dataset_id}/logs', data)
    log_id = ret['result']

    with pytest.raises(Exception):
        s3conn.get_object(Bucket='iceprod2-logs', Key=log_id)

    ret = await client.request('GET', f'/datasets/{dataset_id}/logs/{log_id}')
    assert ret['data'] == data['data']


async def test_rest_logs_s3_dataset_task_get(server):
    client = server(roles=['user'], groups=['users'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    data = {'data': fake_data(2000000), 'dataset_id': dataset_id, 'task_id': 'bar', 'name': 'stdout'}
    ret = await client.request('POST', f'/datasets/{dataset_id}/logs', data)
    log_id = ret['result']

    ret = await client.request('GET', f'/datasets/{dataset_id}/tasks/bar/logs')
    assert 'logs' in ret
    assert len(ret['logs']) == 1
    assert ret['logs'][0]['log_id'] == log_id
    assert data['data'] == ret['logs'][0]['data']

    data = {'data': fake_data(200000), 'dataset_id': dataset_id, 'task_id': 'bar', 'name': 'stdout'}
    ret = await client.request('POST', f'/datasets/{dataset_id}/logs', data)
    log_id = ret['result']

    with pytest.raises(Exception):
        s3conn.get_object(Bucket='iceprod2-logs', Key=log_id)

    ret = await client.request('GET', f'/datasets/{dataset_id}/tasks/bar/logs', {'order': 'asc'})
    assert 'logs' in ret
    assert len(ret['logs']) == 2
    assert ret['logs'][1]['log_id'] == log_id
    assert data['data'] == ret['logs'][1]['data']




#!/usr/bin/env python

import os
import subprocess
import argparse
import json
import glob
import logging
import asyncio
from datetime import datetime
from pprint import pprint

from rest_tools.client import RestClient
from iceprod.core.serialization import dict_to_dataclasses
from iceprod.core.parser import ExpParser

logger = logging.getLogger()

async def run(rest_client, filename, dataset_num=None, job_index=None):
    if not dataset_num:
        # get dataset
        parts = filename.split('/')
        for p in parts[:-1]:
            try:
                dataset_num = int(p)
                if 20000 < dataset_num < 30000:
                    break
            except Exception:
                continue
        else:
            raise Exception('could not determine dataset number')
        logger.info(f'dataset num: {dataset_num}')

    # get dataset_id
    datasets = await rest_client.request('GET', '/datasets?keys=dataset_id|dataset|jobs_submitted')
    for dataset_id in datasets:
        if datasets[dataset_id]['dataset'] == dataset_num:
            jobs_submitted = datasets[dataset_id]['jobs_submitted']
            break
    else:
        raise Exception(f'dataset num {dataset_num} not found')
    logger.info(f'dataset_id: {dataset_id}')

    # get config
    config_url = f'https://iceprod2.icecube.wisc.edu/config?dataset_id={dataset_id}'
    logger.info(f'config_url: {config_url}')
    config = await rest_client.request('GET', f'/config/{dataset_id}')
    config = dict_to_dataclasses(config)

    # get output file patterns
    parser = ExpParser()
    files = []
    for task in config['tasks']:
        for d in task['data']:
            if d['type'] in ('permanent','site_temp') and d['movement'] in ('output','both'):
                files.append({'url': d['remote'], 'iters': 1, 'task': task['name']})
        for tray in task['trays']:
            for d in tray['data']:
                if d['type'] in ('permanent','site_temp') and d['movement'] in ('output','both'):
                    files.append({'url': d['remote'], 'iters': tray['iterations'], 'task': task['name']})
            for module in tray['modules']:
                for d in module['data']:
                    if d['type'] in ('permanent','site_temp') and d['movement'] in ('output','both'):
                        files.append({'url': d['remote'], 'iters': tray['iterations'], 'task': task['name']})

    async def success():
        ret = await rest_client.request('GET', f'/datasets/{dataset_id}/tasks', {'job_index': config['options']['job'], 'keys': 'name|task_id|job_id|task_index'})
        for task in ret.values():
            if task['name'] == config['options']['task']:
                break
        else:
            raise Exception('cannot get task info')
        data = {
            'dataset': dataset_num,
            'dataset_id': dataset_id,
            'job': config['options']['job'],
            'job_id': task['job_id'],
            'task': task['name'],
            'task_id': task['task_id'],
            'config': config_url,
        }
        return data

    env = {'parameters': config['steering']['parameters']}
    config['options'].update({
        'dataset': dataset_num,
        'dataset_id': dataset_id,
        'jobs_submitted': jobs_submitted,
    })
    if job_index:
        job_search = [job_index]
    else:
        job_search = range(jobs_submitted)
    for f in reversed(files):
        logger.info(f'searching task {f["task"]}')
        config['options']['task'] = f['task']
        for j in job_search:
            config['options']['job'] = j
            for i in range(f['iters']):
                config['options']['iter'] = i
                url = parser.parse(f['url'], config, env)
                if '//' not in url:
                    path = url
                else:
                    path = '/'+url.split('//',1)[1].split('/',1)[1]
                logger.info(f'checking path {path}')
                if path == filename:
                    logger.info(f'success on job_index: {j}, iter: {i}')
                    return await success()

    raise Exception('no path match found')

def main():
    parser = argparse.ArgumentParser(description='get IceProd info for an output file')
    parser.add_argument('-t', '--token',help='auth token')
    parser.add_argument('--log_level', default='info', help='log level')
    parser.add_argument('--dataset', type=int, help='dataset num')
    parser.add_argument('--job', type=int, help='job index')
    parser.add_argument('file',help='file')
    
    args = parser.parse_args()
    args = vars(args)

    logging.basicConfig(level=getattr(logging, args['log_level'].upper()))
    
    rpc = RestClient('https://iceprod2-api.icecube.wisc.edu', args['token'])

    ret = asyncio.run(run(rpc, args['file'], dataset_num=args['dataset'], job_index=args['job']))
    pprint(ret)

if __name__ == '__main__':
    main()




import asyncio
import logging
from pathlib import Path
import pkgutil
import time

from rest_tools.client import RestClient, ClientCredentialsAuth

from iceprod.common.mongo_queue import AsyncMongoQueue
from iceprod.services.base import BaseAction, TimeoutException

from .config import get_config
from ..core.logger import stderr_logger

logger = logging.getLogger('service')


async def main() -> None:
    config = get_config()

    message_queue = AsyncMongoQueue(
        url=config.DB_URL,
        collection_name='services_queue',
        timeout=config.DB_TIMEOUT,
        write_concern=config.DB_WRITE_CONCERN
    )

    rest_client: RestClient
    if config.ICEPROD_API_CLIENT_ID and config.ICEPROD_API_CLIENT_SECRET:
        logging.info(f'enabling auth via {config.OPENID_URL} for aud "{config.OPENID_AUDIENCE}"')
        rest_client = ClientCredentialsAuth(
            address=config.ICEPROD_API_ADDRESS,
            token_url=config.OPENID_URL,
            client_id=config.ICEPROD_API_CLIENT_ID,
            client_secret=config.ICEPROD_API_CLIENT_SECRET,
        )
    elif config.CI_TESTING:
        rest_client = RestClient(config.ICEPROD_API_ADDRESS, timeout=1, retries=0)
    else:
        raise RuntimeError('ICEPROD_API_CLIENT_ID or ICEPROD_API_CLIENT_SECRET not specified, and CI_TESTING not enabled!')

    cred_client: RestClient
    if config.ICEPROD_CRED_CLIENT_ID and config.ICEPROD_CRED_CLIENT_SECRET:
        logging.info(f'enabling auth via {config.OPENID_URL} for aud "{config.OPENID_AUDIENCE}"')
        cred_client = ClientCredentialsAuth(
            address=config.ICEPROD_CRED_ADDRESS,
            token_url=config.OPENID_URL,
            client_id=config.ICEPROD_CRED_CLIENT_ID,
            client_secret=config.ICEPROD_CRED_CLIENT_SECRET,
        )
    elif config.CI_TESTING:
        cred_client = RestClient(config.ICEPROD_CRED_ADDRESS, timeout=1, retries=0)
    else:
        raise RuntimeError('ICEPROD_CRED_CLIENT_ID or ICEPROD_CRED_CLIENT_SECRET not specified, and CI_TESTING not enabled!')

    # find all actions
    actions = {}
    plugin_path = str(Path(__file__).parent / 'actions')
    for _, name, _ in pkgutil.iter_modules([plugin_path]):
        action_class = pkgutil.resolve_name(f'iceprod.services.actions.{name}:Action')
        action : BaseAction = action_class(queue=message_queue, logger=logger, api_client=rest_client, cred_client=cred_client)
        actions[name] = action

    # process messages
    while True:
        sleep_time = 0.
        try:
            start_time = time.time()

            data = await message_queue.pop(timeout_seconds=config.SERVICE_TIMEOUT+10)

            if data:
                try:
                    type_ = data.payload.pop('type')
                    logging.info('running service for type %s', type_)
                    fut = actions[type_].run(data)
                    await asyncio.wait_for(fut, timeout=config.SERVICE_TIMEOUT)
                except TimeoutException:
                    await message_queue.release(data.uuid)
                except Exception as e:
                    await message_queue.fail(data.uuid, error_message=str(e))
                    raise
                else:
                    await message_queue.complete(data.uuid)
            else:
                logging.info('service has nothing to do. sleeping')
                sleep_time = config.SERVICE_SLEEP_SECS - (time.time() - start_time)
        except Exception:
            logging.error('error running services', exc_info=True)

        if config.CI_TESTING:
            break  # only process one message during testing

        if sleep_time > 0:
            await asyncio.sleep(sleep_time)

    await message_queue.close()

if __name__ == '__main__':
    stderr_logger()
    asyncio.run(main())



import functools
import logging
import re
import time
import traceback
from typing import Any
from urllib.parse import urlencode

from iceprod.core.jsonUtil import json_encode

import tornado.web
import jwt
import requests.exceptions
from rest_tools.client import RestClient, OpenIDRestClient
from rest_tools.server import catch_error, RestHandler
from rest_tools.server.session import SessionMixin

from iceprod.credentials.util import get_expiration
from iceprod.util import VERSION_STRING
from iceprod.common.prom_utils import PromRequestMixin
from iceprod.roles_groups import GROUPS
import iceprod.server.states

logger = logging.getLogger('website_base')


def authenticated(method):
    """Decorate methods with this to require that the user be logged in.

    If the user is not logged in, they will be redirected to the configured
    `login url <RequestHandler.get_login_url>`.

    If you configure a login url with a query parameter, Tornado will
    assume you know what you're doing and use it as-is.  If not, it
    will add a `next` parameter so the login page knows where to send
    you once you're logged in.
    """
    @catch_error
    @functools.wraps(method)
    async def wrapper(self, *args, **kwargs):
        if not await self.get_current_user_async():
            if self.request.method in ("GET", "HEAD"):
                try:
                    url = self.get_login_url()
                    if "?" not in url:
                        next_url = self.request.uri
                        url += "?" + urlencode(dict(next=next_url))
                    self.redirect(url)
                    return None
                except Exception:
                    logger.warning('failed to make redirect', exc_info=True)
                    raise tornado.web.HTTPError(403, reason='auth failed')
            raise tornado.web.HTTPError(403, reason='auth failed')
        return await method(self, *args, **kwargs)
    return wrapper


def eval_expression(token, e):
    """from rest_tools.server.decorators.token_attribute_role_mapping_auth"""
    name, val = e.split('=',1)
    if name == 'scope':
        # special handling to split into string
        token_val = token.get('scope','').split()
    else:
        prefix = name.split('.')[:-1]
        while prefix:
            token = token.get(prefix[0], {})
            prefix = prefix[1:]
        token_val = token.get(name.split('.')[-1], None)

    logger.debug('token_val = %r', token_val)
    if token_val is None:
        return []

    prog = re.compile(val)
    if isinstance(token_val, list):
        ret = (prog.fullmatch(v) for v in token_val)
    else:
        ret = [prog.fullmatch(token_val)]
    return [r for r in ret if r]


class LoginMixin(SessionMixin, RestHandler):  # type: ignore[misc]
    """
    Store/load current user's `OpenIDLoginHandler` tokens in Redis.
    """
    def get_current_user(self) -> str | None:
        """Get the current username from the cookie"""
        try:
            username = self.get_secure_cookie('iceprod_username')
            if not username:
                raise RuntimeError('missing iceprod_username cookie')
            return username.decode('utf-8')
        except Exception:
            logger.info('failed to get username', exc_info=True)
        return None

    @property
    def auth_access_token(self) -> str | None:
        assert self.auth
        if self.session:
            ret = self.session.get('access_token', None)
            if not isinstance(ret, str):
                logger.info('bad access token type: not str')
                return None
            return ret
        else:
            logger.info('no session to get access token from')
        return None

    @auth_access_token.setter
    def auth_access_token(self, val: str):
        if self.session:
            self.session['access_token'] = val
        else:
            raise RuntimeError('no valid session')

    @property
    def auth_refresh_token(self) -> str | None:
        if self.session:
            ret = self.session.get('refresh_token', None)
            if not isinstance(ret, str):
                logger.info('bad access token type: not str')
                return None
            return ret
        else:
            logger.info('no session to get refresh token from')
        return None

    @auth_refresh_token.setter  # type: ignore[override]
    def auth_refresh_token(self, val: str):
        if self.session:
            self.session['refresh_token'] = val
        else:
            raise RuntimeError('no valid session')

    @property
    def auth_groups(self) -> list[str]:
        assert self.auth
        if not self.auth_access_token:
            return []
        try:
            data = self.auth.validate(self.auth_access_token)
        except jwt.ExpiredSignatureError:
            logger.debug('user access_token expired')
            return []

        # lookup groups
        groups: set[str] = set()
        try:
            for name in GROUPS:
                for expression in GROUPS[name]:
                    ret = eval_expression(data, expression)
                    groups.update(match.expand(name) for match in ret)
        except Exception:
            logger.info('cannot determine groups', exc_info=True)
        return sorted(groups)

    def store_tokens(
        self,
        access_token,
        access_token_exp,
        refresh_token=None,
        refresh_token_exp=None,
        user_info=None,
        user_info_exp=None,
    ):
        """
        Store jwt tokens and user info from OpenID-compliant auth source.

        Args:
            access_token (str): jwt access token
            access_token_exp (int): access token expiration in seconds
            refresh_token (str): jwt refresh token
            refresh_token_exp (int): refresh token expiration in seconds
            user_info (dict): user info (from id token or user info lookup)
            user_info_exp (int): user info expiration in seconds
        """
        assert self.auth
        if not user_info:
            user_info = self.auth.validate(access_token)
        username = user_info.get('preferred_username')
        if not username:
            username = user_info.get('upn')
        if not username:
            raise tornado.web.HTTPError(400, reason='no username in token')

        data = {
            'access_token': access_token,
            'refresh_token': refresh_token,
        }
        self._session_mgr.set(username, data)

        self.set_secure_cookie('iceprod_username', username, expires_days=30)

    def clear_tokens(self):
        """
        Clear token data, usually on logout.
        """
        self.clear_cookie('iceprod_username')
        if username := self.current_user:
            self._session_mgr.delete_session(username)


class TokenStorageMixin(RestHandler):
    """
    Store/load current user's tokens in iceprod credentials API.
    """
    TokenResult = list[dict[str, Any]]

    def initialize(self, *args, cred_rest_client, **kwargs):
        super().initialize(**kwargs)
        self.cred_rest_client = cred_rest_client

    async def _get_cred_tokens(self, path: str, url: str | None = None, scope: str | None = None) -> TokenResult:
        """Get selected tokens from the credential service."""
        try:
            assert self.auth
            args = {'url': url}
            if scope:
                args['scope'] = scope
            creds = await self.cred_rest_client.request('GET', path, args)
            return creds
        except requests.exceptions.RequestException:
            logger.warning('failed to get credentials', exc_info=True)
        return []

    async def get_cred_group_tokens(self, group_name: str, url: str | None = None, scope: str | None = None) -> TokenResult:
        return await self._get_cred_tokens(
            path=f'/groups/{group_name}/credentials',
            url=url,
            scope=scope,
        )

    async def get_cred_user_tokens(self, username: str, url: str | None = None, scope: str | None = None) -> TokenResult:
        return await self._get_cred_tokens(
            path=f'/users/{username}/credentials',
            url=url,
            scope=scope,
        )

    async def get_cred_dataset_tokens(self, dataset_id: str, url: str | None = None, scope: str | None = None) -> TokenResult:
        return await self._get_cred_tokens(
            path=f'/datasets/{dataset_id}/credentials',
            url=url,
            scope=scope,
        )

    async def get_cred_dataset_task_tokens(self, dataset_id: str, task_name: str, url: str | None = None, scope: str | None = None) -> TokenResult:
        return await self._get_cred_tokens(
            path=f'/datasets/{dataset_id}/tasks/{task_name}/credentials',
            url=url,
            scope=scope,
        )

    async def _put_cred_tokens(
        self,
        path: str,
        url: str,
        access_token: str,
        refresh_token: str | None = None
    ):
        """
        Store jwt tokens from OpenID-compliant auth source.

        Args:
            url (str): site url
            access_token (str): jwt access token
            refresh_token (str): jwt refresh token
        """
        assert self.auth
        args = {
            'url': url,
            'type': 'oauth',
            'access_token': access_token,
        }
        if refresh_token:
            args['refresh_token'] = refresh_token

        await self.cred_rest_client.request('POST', path, args)

    async def put_cred_group_tokens(
            self,
            group_name: str,
            url: str,
            access_token: str,
            refresh_token: str | None = None
    ):
        return await self._put_cred_tokens(
            path=f'/groups/{group_name}/credentials',
            url=url,
            access_token=access_token,
            refresh_token=refresh_token,
        )

    async def put_cred_user_tokens(
            self,
            username: str,
            url: str,
            access_token: str,
            refresh_token: str | None = None
    ):
        return await self._put_cred_tokens(
            path=f'/users/{username}/credentials',
            url=url,
            access_token=access_token,
            refresh_token=refresh_token,
        )

    async def put_cred_dataset_task_tokens(
            self,
            dataset_id: str,
            task_name: str,
            url: str,
            access_token: str,
            refresh_token: str | None = None
    ):
        return await self._put_cred_tokens(
            path=f'/datasets/{dataset_id}/tasks/{task_name}/credentials',
            url=url,
            access_token=access_token,
            refresh_token=refresh_token,
        )

    async def _clear_cred_tokens(self, path: str, url: str | None = None, scope: str | None = None):
        """
        Clear all token data.
        """
        args = {}
        if url:
            args['url'] = url
        if scope:
            args['scope'] = scope
        await self.cred_rest_client.request('DELETE', path, args)

    async def clear_cred_group_tokens(self, group_name: str, url: str, scope: str | None = None):
        return await self._clear_cred_tokens(
            path=f'/groups/{group_name}/credentials',
            url=url,
            scope=scope,
        )

    async def clear_cred_user_tokens(self, username: str, url: str, scope: str | None = None):
        return await self._clear_cred_tokens(
            path=f'/users/{username}/credentials',
            url=url,
            scope=scope,
        )

    async def clear_cred_dataset_tokens(self, dataset_id: str, url: str, scope: str | None = None):
        return await self._clear_cred_tokens(
            path=f'/datasets/{dataset_id}/credentials',
            url=url,
            scope=scope,
        )

    async def clear_cred_dataset_task_tokens(self, dataset_id: str, task_name: str, url: str, scope: str | None = None):
        return await self._clear_cred_tokens(
            path=f'/datasets/{dataset_id}/tasks/{task_name}/credentials',
            url=url,
            scope=scope,
        )


class PublicHandler(LoginMixin, TokenStorageMixin, PromRequestMixin, RestHandler):  # type: ignore[override]
    """Default Handler"""
    def initialize(  # type: ignore[override]
        self,
        *args,
        rest_api: str,
        system_rest_client: RestClient,
        auth_url: str,
        auth_client_id: str,
        auth_client_secret: str | None = None,
        **kwargs
    ):
        """
        Get some params from the website module

        :param rest_api: the rest api url
        :param system_rest_client: the rest client for the system role
        """
        super().initialize(*args, **kwargs)
        self.rest_api = rest_api
        self.system_rest_client = system_rest_client
        self.rest_client: RestClient | None = None
        self.auth_url = auth_url
        self.auth_client_id = auth_client_id
        self.auth_client_secret = auth_client_secret

    def get_template_namespace(self) -> dict[str, Any]:
        namespace = super().get_template_namespace()
        namespace['version'] = VERSION_STRING
        if self.request.uri:
            namespace['section'] = self.request.uri.lstrip('/').split('?')[0].split('/')[0]
        namespace['json_encode'] = json_encode
        namespace['states'] = iceprod.server.states
        namespace['rest_api'] = self.rest_api
        return namespace

    def update_refresh_token(self, access: str | bytes, refresh: str | bytes | None):
        if access:
            if isinstance(access, bytes):
                access = access.decode('utf-8')
            self.auth_access_token = access  # type: ignore
        if refresh:
            if isinstance(refresh, bytes):
                refresh = refresh.decode('utf-8')
            self.auth_refresh_token = refresh  # type: ignore

    async def get_current_user_async(self) -> str | None:
        try:
            self.current_user = LoginMixin.get_current_user(self)
            if self.current_user and self.auth_refresh_token:
                self.rest_client = OpenIDRestClient(
                    address=self.rest_api,
                    token_url=self.auth_url,
                    refresh_token=self.auth_refresh_token,
                    client_id=self.auth_client_id,
                    client_secret=self.auth_client_secret,
                    update_func=self.update_refresh_token,
                    timeout=50,
                    retries=1,
                )
                # verify refresh works  (note: this contacts Keycloak every time!)
                if not self.rest_client._openid_token():
                    logger.info('cannot refresh token')
                    return None
            elif self.current_user and self.auth_access_token:
                if get_expiration(self.auth_access_token) <= time.time():
                    logger.info('access token expired')
                    return None
                self.rest_client = RestClient(self.rest_api, self.auth_access_token, timeout=50, retries=1)
            elif not self.current_user:
                logger.info('no current user')
                return None
            else:
                logger.info('no access token')
                return None
        except Exception:
            logger.info('failed to create user rest client', exc_info=True)
            return None
        return self.current_user

    def write_error(self, status_code: int = 500, **kwargs) -> None:
        """Write out custom error page."""
        self.set_status(status_code)
        if status_code >= 500:
            self.write('<h2>Internal Error</h2>')
        else:
            self.write('<h2>Request Error</h2>')
        if 'exc_info' in kwargs:
            exception = kwargs["exc_info"][1]
            logger.info(''.join(traceback.format_exception(exception)))
            if isinstance(exception, tornado.web.HTTPError) and exception.reason:
                kwargs['reason'] = exception.reason
        if 'reason' in kwargs:
            self.write('<br />'.join(kwargs['reason'].split('\n')))
        elif 'message' in kwargs:
            self.write('<br />'.join(kwargs['message'].split('\n')))
        elif self._reason:
            self.write('<br />'.join(self._reason.split('\n')))
        self.finish()



"""
Test script for scheduled_tasks/dataset_monitor
"""

from collections import defaultdict
import logging
from unittest.mock import AsyncMock, MagicMock

import pytest
from iceprod.scheduled_tasks import dataset_monitor

logger = logging.getLogger('scheduled_tasks_dataset_monitor_test')


async def test_200_run():
    rc = MagicMock()
    pilots = {}
    jobs = {}
    tasks = defaultdict(dict)
    stats = {}
    async def client(method, url, args=None):
        logger.info('REST: %s, %s', method, url)
        if url.startswith('/dataset_summaries/status'):
            return {'processing':['foo']}
        elif url.startswith('/datasets/foo/job_counts/status'):
            return jobs
        elif url.startswith('/datasets/foo/task_counts/name_status'):
            client.called = True
            return tasks
        elif url.startswith('/datasets/foo/task_stats'):
            return stats
        if url.startswith('/datasets/foo'):
            return {'dataset':123,'status':'processing','jobs_submitted':1,'tasks_submitted':1}
        else:
            raise Exception()
    client.called = False
    rc.request = client

    await dataset_monitor.run(rc, debug=True)
    assert client.called

    jobs['processing'] = 1
    await dataset_monitor.run(rc, debug=True)

    tasks['generate']['queued'] = 1
    await dataset_monitor.run(rc, debug=True)


async def test_201_run():
    rc = MagicMock()
    rc.request = AsyncMock(side_effect=Exception())
    with pytest.raises(Exception):
        await dataset_monitor.run(rc, debug=True)

    # check it normally hides the error
    await dataset_monitor.run(rc, debug=False)



"""
Make a call graph for async functions.

Does not catch regular function calls.
"""
import string
from collections import OrderedDict

# utils
def is_func(line):
    line = line.strip(string.whitespace)
    return line.startswith('async def') or line.startswith('def')
all_whitespace = lambda line: not line.strip(string.whitespace)
def get_indent(line):
    for i,c in enumerate(line):
        if c not in string.whitespace:
            return i
    return -1

def find_func_names(lines):
    """Find all function names"""
    def get_name(line):
        return line.split('def ',1)[-1].split('(',1)[0]
    names = []
    found = False
    indent = -1
    for line in lines:
        line = line.strip('\r\n')
        if found and not all_whitespace(line):
            if get_indent(line) <= indent:
                found = False
        if (not found) and is_func(line):
            found = True
            indent = get_indent(line)
            names.append(get_name(line))
    return names

def find_func(lines, name):
    """Find the lines of a specific function"""
    is_my_func = lambda line: is_func(line) and name in line.split('(')[0]

    found = False
    indent = -1
    last_line = ''
    for line in lines:
        line = line.strip('\r\n')
        if last_line:
            line = last_line+line.strip()
        if line.strip().endswith(','):
            last_line = line
        else:
            last_line = ''
        if (not found) and is_my_func(line):
            found = True
            indent = get_indent(line)
        elif found and not all_whitespace(line):
            if get_indent(line) > indent:
                yield line
            else:
                return # end of function
    if not found:
        raise Exception(f'{name} not found')

def process_func(lines, func_names):
    """Search for function calls"""
    ret = OrderedDict()
    for line in lines:
        #print(f':: {line}')
        for n in func_names:
            if n+'(' in line:
                name = line.split(n+'(')[0].split('(')[-1].split()[-1]+n
                if name in ret:
                    ret[name] += 1
                else:
                    ret[name] = 1
        if 'await' in line:
            line = line.split('await',1)[-1].strip()
            if line.startswith('asyncio.ensure_future'):
                line = line.split('(',1)[-1]
            if 'rest_client.request' in line:
                line = line.split(')',1)[0]+')'
            else:
                line = line.split('(',1)[0]
            if line in ret:
                ret[line] += 1
            else:
                ret[line] = 1
    return ret

def analyze_calls(lines, funcname, indent=0, recurse=True):
    func_names = find_func_names(lines)
    calls = process_func(find_func(lines, funcname), func_names)
    for c in calls:
        if '.' in c and not c.startswith('self'):
            continue
        print(' '*indent+c)
        if recurse:
            if c.startswith('self.'):
                c = c[5:]
            try:
                analyze_calls(lines, c, indent=indent+2, recurse=True)
            except Exception as e:
                #print(' '*indent,e)
                pass

def main():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('filename')
    parser.add_argument('function')
    parser.add_argument('-r','--recurse',default=True)
    args = parser.parse_args()

    print(f'searching for {args.function} in file {args.filename}')
    with open(args.filename) as f:
        lines = f.readlines()

    print('')
    print('Calls: ')
    analyze_calls(lines, args.function, indent=0, recurse=args.recurse)

if __name__ == '__main__':
    main()


import os
import sys
sys.path.insert(0, os.path.abspath('..'))
import iceprod
from iceprod.util import VERSION_STRING

# the dbmethod decorator and docstrings don't get along
# so turn the decorator off
def nodecorate(func):
    return func
#import iceprod.server.dbmethods
#iceprod.server.dbmethods.dbmethod = nodecorate

master_doc = "index"

project = 'IceProd'
copyright = '2016, IceCube Collaboration'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
is_master = False
try:
    with open('../.git/HEAD') as f:
        is_master = 'master' in f.read()
except:
    pass
if is_master:
    release = 'master'
    version = VERSION_STRING
else:
    release = VERSION_STRING
    version = VERSION_STRING

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.extlinks',
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# -- Autodoc options -----------------------------------------------------------------

autodoc_member_order = 'bysource'
autoclass_content = 'both'
autodoc_default_options = {
  'members': True,
  'undoc-members': True,
}

# Without this line sphinx includes a copy of object.__init__'s docstring
# on any class that doesn't define __init__.
# https://bitbucket.org/birkenfeld/sphinx/issue/1337/autoclass_content-both-uses-object__init__
autodoc_docstring_signature = False

coverage_skip_undoc_in_source = True
coverage_ignore_modules = [
    'iceprod.core.tests'
    'iceprod.server.tests',
    ]
coverage_ignore_classes = [
    ]

coverage_ignore_functions = [
    # various modules
    "test",
    "main",
]

# -- Options for HTML output ---------------------------------------------------

html_theme = '_theme'
html_theme_path = ['.']

#html_favicon = None

# -- Options for LaTeX output --------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ('index', 'IceProd.tex', u'IceProd Documentation',
   u'David Schultz', 'manual'),
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3.6/', None),
    'tornado': ('http://tornado.readthedocs.io/en/stable/', None),
    }



"""
Test script for main iceprod server
"""

from __future__ import absolute_import, division, print_function

from tests.util import unittest_reporter, glob_tests

import logging
logger = logging.getLogger('iceprod_server_test')

import os
import time
import tempfile
import shutil
import subprocess

import unittest

import psutil

def get_children():
    ret = {}
    for p in psutil.process_iter():
        try:
            attrs = p.as_dict(['name','cmdline'])
            if 'iceprod' in attrs['name'] or 'iceprod' in str(attrs['cmdline']):
                ret[p.pid()] = p
                for pp in p.children(recursive=True):
                    ret[pp.pid()] = pp
        except:
            pass
    if os.getpid() in ret:
        del ret[os.getpid()]
    return ret.values()

class iceprod_server_test(unittest.TestCase):
    def setUp(self):
        super(iceprod_server_test,self).setUp()

        self.test_dir = tempfile.mkdtemp(dir=os.getcwd())
        curdir = os.getcwd()
        os.chdir(self.test_dir)
        def cleanup():
            os.chdir(curdir)
            shutil.rmtree(self.test_dir)
        self.addCleanup(cleanup)

    @unittest_reporter(skip=True)
    def test_01_run(self):
        # make sure we don't try running the grid
        with open('iceprod_config.json','w') as f:
            f.write('{"modules":{"queue":false},"logging":{"level":"debug"}}')
        subprocess.check_call(['../bin/iceprod_server.py','--pidfile','pid','start'])
        try:
            time.sleep(4)
            self.assertTrue(os.path.exists('pid'))
            alive_children = get_children()
            logger.info('children: %r', alive_children)

            subprocess.check_call(['../bin/iceprod_server.py','--pidfile','pid','stop'])
            self.assertFalse(os.path.exists('pid'))
            dead_children = get_children()
            if dead_children:
                raise Exception('did not kill all children')
        finally:
            if os.path.exists('var/log'):
                for f in os.listdir('var/log'):
                    p = os.path.join('var/log',f)
                    if os.path.isfile(p):
                        logger.info('log for %s:\n\n%s\n\n', f, open(p).read())
            if os.path.exists('pid'):
                subprocess.check_call(['../bin/iceprod_server.py','--pidfile','pid','kill'])

def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    alltests = glob_tests(loader.getTestCaseNames(iceprod_server_test))
    suite.addTests(loader.loadTestsFromNames(alltests,iceprod_server_test))
    return suite



#!/usr/bin/env python

import os
import subprocess
import argparse
import json
import glob
import tempfile
import shutil
import logging
from contextlib import contextmanager

from rest_tools.client import RestClient

def cleanup():
    for file in glob.glob('iceprod_*'):
        os.remove(file)
    try:
        shutil.rmtree('local_temp')
    except Exception:
        pass

def run(token, config, jobs_submitted, job, task, clean=True, debug=False):
    cleanup()
    cmd = ['python', '-m', 'iceprod.core.i3exec', '-p', token,
           '-u', 'https://iceprod2-api.icecube.wisc.edu',
           '--jobs_submitted', f'{jobs_submitted}', '-f', config,
           '--job', f'{job}', '--task', f'{task}']
    if debug:
        cmd += ['--debug']
    subprocess.check_call(cmd)
    if clean:
        cleanup()

@contextmanager
def make_pilot(rpc):
    resources = {'cpu': 1, 'gpu': 0, 'memory': 4, 'disk': 1000, 'time': 24}
    pilot = {'resources': resources,
             'resources_available': resources,
             'resources_claimed': resources,
             'tasks': [],
             'queue_host': 'manual_run',
             'queue_version': 'unknown',
             'version': 'unknown',
    }
    ret = rpc.request_seq('POST', '/pilots', pilot)
    pilot_id = ret['result']

    def set_task(task_id):
        data = {'tasks': [task_id]}
        rpc.request_seq('PATCH', f'/pilots/{pilot_id}', data)

    try:
        yield set_task
    finally:
        rpc.request_seq('DELETE', f'/pilots/{pilot_id}')

def write_config(config, filename, dataset_id, dataset, task_id, subprocess_dir=os.getcwd()):
    data = config.copy()
    data['dataset'] = dataset
    data['options'] = {
        'dataset_id': dataset_id,
        'dataset': dataset,
        'task_id': task_id,
        'subprocess_dir': subprocess_dir,
    }
    with open(filename, 'w') as f:
        json.dump(data, f)

def main():
    parser = argparse.ArgumentParser(description='manually run IceProd i3exec')
    parser.add_argument('-t', '--token', help='auth token')
    parser.add_argument('-d','--dataset', type=int, help='dataset number')
    parser.add_argument('-j','--job', type=int, help='job number (optional)')
    parser.add_argument('--run-failed-jobs', action='store_true', help='also run failed jobs')
    parser.add_argument('--no-clean', dest='clean', action='store_false', help='do not clean up after job')
    parser.add_argument('--log-level', default='DEBUG', choices=['ERROR','WARNING','INFO','DEBUG'], help='log level')
    parser.add_argument('--ignore-error', action='store_true', help='keep going if a job fails')
    
    args = parser.parse_args()
    args = vars(args)

    logging.basicConfig(level=args['log_level'])

    rpc = RestClient('https://iceprod2-api.icecube.wisc.edu', args['token'])
    
    datasets = rpc.request_seq('GET', '/datasets', {'keys': 'dataset_id|dataset'})
    dataset_id = None
    for d in datasets:
        if datasets[d]['dataset'] == args['dataset']:
            dataset_id = d
            break
    else:
        raise Exception('bad dataset num')
    dataset = rpc.request_seq('GET', f'/datasets/{dataset_id}')
    config = rpc.request_seq('GET', f'/config/{dataset_id}')

    jobs = rpc.request_seq('GET', f'/datasets/{dataset_id}/jobs', {'status': 'processing|errors'})
    if args['job'] is not None:
        jobs = {j:jobs[j] for j in jobs if jobs[j]['job_index'] == args['job']}
    if not jobs:
        raise Exception('no jobs found')

    with make_pilot(rpc) as pilot:
        for job_id in jobs:
            status = 'waiting|queued|reset'
            if args['run_failed_jobs']:
                status += '|failed'
            tasks = rpc.request_seq('GET', f'/datasets/{dataset_id}/tasks',
                                    {'job_id': job_id, 'keys': 'task_id|task_index|name|depends',
                                     'status': status})
            if not tasks:
                logging.warning(f'no tasks available for {dataset["dataset"]} {jobs[job_id]["job_index"]}')
            for task_id in sorted(tasks, key=lambda t:tasks[t]['task_index']):
                ret = rpc.request_seq('GET', f'/datasets/{dataset_id}/tasks/{task_id}')
                if ret['status'] not in status.split('|'):
                    continue
                print(f'processing {dataset["dataset"]} {jobs[job_id]["job_index"]} {tasks[task_id]["name"]}')

                tmpdir = tempfile.mkdtemp(suffix='.{}'.format(task_id), dir=os.getcwd())
                try:
                    write_config(config, 'config.json', dataset_id, args['dataset'], task_id, subprocess_dir=tmpdir)
                    pilot(task_id)
                    run(token=args['token'], config='config.json',
                        jobs_submitted=dataset['jobs_submitted'],
                        job=jobs[job_id]['job_index'],
                        task=tasks[task_id]['name'],
                        clean=args['clean'],
                        debug=args['log_level'] == 'DEBUG')
                except subprocess.SubprocessError:
                    if args['ignore_error']:
                        logging.warn("error in subprocess", exc_info=True)
                        continue
                    raise
                finally:
                    shutil.rmtree(tmpdir)

    try:
        os.remove('config.json')
    except Exception:
        pass

if __name__ == '__main__':
    main()



"""
Test script for scheduled_tasks/job_temp_cleaning
"""

import logging
from datetime import datetime, timedelta, UTC
from unittest.mock import MagicMock, AsyncMock

import pytest
from iceprod.scheduled_tasks import job_temp_cleaning

logger = logging.getLogger('scheduled_tasks_job_temp_cleaning_test')


async def test_scheduled_tasks_job_temp_cleaning_run():
    rc = MagicMock()
    jobs = {}

    async def client(method, url, args=None):
        logger.info('REST: %s, %s', method, url)
        if url.startswith('/datasets?'):
            return {'0':{'dataset_id':'0', 'dataset':0}}
        elif url.startswith('/datasets/0/jobs'):
            return jobs
        else:
            raise Exception()
    rc.request = AsyncMock(side_effect=client)

    listdir = AsyncMock()
    rmtree = AsyncMock()

    path = '/foo'
    data = {'0': {'1': 1024}}

    # empty dir
    listdir.return_value = {}
    await job_temp_cleaning.run(rc, path, listdir, rmtree, debug=True)
    assert rc.request.await_count == 1
    listdir.assert_awaited()
    rmtree.assert_not_awaited()

    # dir with job, but no job in db
    rc.request = AsyncMock(side_effect=client)
    listdir = AsyncMock(return_value=data)
    await job_temp_cleaning.run(rc, path, listdir, rmtree, debug=True)
    listdir.assert_awaited()
    assert rc.request.await_count == 2
    rmtree.assert_not_awaited()

    # dir with completed job
    jobs['bar'] = {'job_index': 1, 'status': 'complete'}
    rc.request = AsyncMock(side_effect=client)
    listdir = AsyncMock(return_value=data)
    await job_temp_cleaning.run(rc, path, listdir, rmtree, debug=True)
    assert rc.request.await_count == 2
    listdir.assert_awaited()
    rmtree.assert_awaited_once_with(path+'/0/1')

    # dir with recent suspended job
    now = datetime.now(UTC)
    jobs['bar'] = {
        'job_index': 1,
        'status': 'suspended',
        'status_changed': now.strftime('%Y-%m-%dT%H:%M:%S.%f'),
    }
    rc.request = AsyncMock(side_effect=client)
    listdir = AsyncMock(return_value=data)
    rmtree.reset_mock()
    await job_temp_cleaning.run(rc, path, listdir, rmtree, debug=True)
    assert rc.request.await_count == 2
    listdir.assert_awaited()
    rmtree.assert_not_awaited()

    # dir with old suspended job
    now = datetime.now(UTC)
    jobs['bar'] = {
        'job_index': 1,
        'status': 'suspended',
        'status_changed': (now-timedelta(days=100)).strftime('%Y-%m-%dT%H:%M:%S'),
    }
    rc.request = AsyncMock(side_effect=client)
    listdir = AsyncMock(return_value=data)
    await job_temp_cleaning.run(rc, path, listdir, rmtree, debug=True)
    assert rc.request.await_count == 2
    listdir.assert_awaited()
    rmtree.assert_awaited_once_with(path+'/0/1')

    # error in rmtree
    rc.request = AsyncMock(side_effect=client)
    listdir = AsyncMock(return_value=data)
    rmtree = AsyncMock(side_effect=Exception('bad rmtree'))
    with pytest.raises(Exception, match='bad rmtree'):
        await job_temp_cleaning.run(rc, path, listdir, rmtree, debug=True)
    assert rc.request.await_count == 2
    listdir.assert_awaited()
    rmtree.assert_awaited_once_with(path+'/0/1')

    # error in listdir
    rc.request = AsyncMock(side_effect=client)
    listdir = AsyncMock(side_effect=Exception('bad listdir'))
    rmtree = AsyncMock()
    with pytest.raises(Exception, match='bad listdir'):
        await job_temp_cleaning.run(rc, path, listdir, rmtree, debug=True)
    assert rc.request.await_count == 0
    listdir.assert_awaited()
    rmtree.assert_not_awaited()



"""The IceProd Grid Middleware and tools."""

from __future__ import absolute_import, division, print_function

# NOTE: `__version__` is not defined because this package is built using 'setuptools-scm' --
#   use `importlib.metadata.version(...)` if you need to access version info at runtime.



import os
import socket

import boto3
from moto import mock_aws
from pymongo import AsyncMongoClient
import pytest
import pytest_asyncio
from wipac_dev_tools import from_environment


@pytest.fixture
def port():
    """Get an ephemeral port number."""
    # https://unix.stackexchange.com/a/132524
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('', 0))
    addr = s.getsockname()
    ephemeral_port = addr[1]
    s.close()
    return ephemeral_port


@pytest_asyncio.fixture
async def mongo_clear():
    default_config = {
        'DB_URL': 'mongodb://localhost/iceprod',
        'DATABASES': '',
    }
    config = from_environment(default_config)
    db_url, db_name = config['DB_URL'].rsplit('/', 1)  # type: ignore
    client = AsyncMongoClient(db_url, serverSelectionTimeoutMS=10)

    async def _clean(db_name):
        db = client[db_name]
        if db_name == 'config':
            # exclude replica set config
            await db['config'].drop()
        else:
            cols = await db.list_collection_names()
            for c in cols:
                await db[c].drop()

    if not config['DATABASES']:
        await _clean(db_name)
    else:
        for db_name in config['DATABASES'].split('|'):  # type: ignore
            await _clean(db_name)


@pytest.fixture(scope='module')
def monkeymodule():
    with pytest.MonkeyPatch.context() as mp:
        yield mp


@pytest.fixture(scope='module')
def mongo_url(monkeymodule):
    if 'DB_URL' not in os.environ:
        monkeymodule.setenv('DB_URL', 'mongodb://localhost/iceprod')


@pytest.fixture
def s3conn(monkeypatch):
    monkeypatch.setenv('S3_ADDRESS', 'http://localhost:5000')
    monkeypatch.setenv('S3_ACCESS_KEY', 'XXXX')
    monkeypatch.setenv('S3_SECRET_KEY', 'XXXX')

    with mock_aws():
        conn = boto3.client('s3', region_name='us-east-1')
        conn.create_bucket(Bucket='iceprod2-logs')
        yield conn


"""
Common functions
"""

from __future__ import absolute_import, division, print_function

import os
import shutil
import logging
import socket
import subprocess
import hashlib
from functools import partial, reduce
from contextlib import contextmanager
import asyncio

try:
    import psutil
except ImportError:
    psutil = None  # type: ignore

import requests
from requests_toolbelt.multipart.encoder import MultipartEncoder  # type: ignore
from rest_tools.client import Session, AsyncSession

from iceprod.core.gridftp import GridFTP


# Compression Functions #
_compress_suffixes = ('.tgz','.gz','.tbz2','.tbz','.bz2','.bz',
                      '.lzma2','.lzma','.lz','.xz')
_tar_suffixes = ('.tar', '.tar.gz', '.tgz', '.tar.bz2', '.tbz2', '.tbz',
                 '.tar.lzma', '.tar.xz', '.tlz', '.txz')


def uncompress(infile, out_dir=None):
    """Uncompress a file, if possible"""
    files = []
    cur_dir = os.getcwd()
    try:
        if out_dir:
            os.chdir(out_dir)
        logging.info('uncompressing %s',infile)
        if istarred(infile):
            # handle tarfile
            output = subprocess.check_output(['tar','-atf',infile]).decode('utf-8')
            files = [x for x in output.split('\n') if x.strip() and x[-1] != '/']
            if not files:
                raise Exception('no files inside tarfile')
            for f in files:
                if os.path.exists(f):
                    break
            else:
                subprocess.call(['tar','-axf',infile])
        else:
            if infile.endswith('.gz'):
                cmd = 'gzip'
            elif any(infile.endswith(s) for s in ('.bz','.bz2')):
                cmd = 'bzip2'
            elif any(infile.endswith(s) for s in ('.xz','.lzma')):
                cmd = 'xz'
            else:
                logging.info('unknown format: %s',infile)
                raise Exception('unknown format')
            subprocess.call([cmd,'-kdf',infile])
            files.append(infile.rsplit('.',1)[0])
    finally:
        os.chdir(cur_dir)

    logging.info('files: %r', files)
    if len(files) == 1:
        return files[0]
    else:
        return files


def compress(infile,compression='lzma'):
    """Compress a file or directory.
       The compression argument is used as the new file extension"""
    if not istarred('.'+compression) and os.path.isdir(infile):
        outfile = infile+'.tar.'+compression
    else:
        outfile = infile+'.'+compression
    if istarred(outfile):
        dirname, filename = os.path.split(infile)
        subprocess.call(['tar','-acf',outfile,'-C',dirname,filename])
    else:
        if outfile.endswith('.gz'):
            cmd = ['gzip']
        elif any(outfile.endswith(s) for s in ('.bz','.bz2')):
            cmd = ['bzip2']
        elif outfile.endswith('.xz'):
            cmd = ['xz']
        elif outfile.endswith('.lzma'):
            cmd = ['xz','-F','lzma']
        else:
            logging.info('unknown format: %s',infile)
            raise Exception('unknown format')
        subprocess.call(cmd+['-kf',infile])
    return outfile


def iscompressed(infile):
    """Check if a file is a compressed file, based on file name"""
    return any(infile.endswith(s) for s in _compress_suffixes)


def istarred(infile):
    """Check if a file is a tarred file, based on file name"""
    return any(infile.endswith(s) for s in _tar_suffixes)


def cksm(filename,type,buffersize=16384,file=True):
    """Return checksum of file using algorithm specified"""
    if type not in ('md5','sha1','sha256','sha512'):
        raise Exception('cannot get checksum for type %r',type)

    try:
        digest = getattr(hashlib,type)()
    except Exception:
        raise Exception('cannot get checksum for type %r',type)

    if file and os.path.exists(filename):
        # checksum file contents
        with open(filename,'rb') as filed:
            buffer = filed.read(buffersize)
            while buffer:
                digest.update(buffer)
                buffer = filed.read(buffersize)
    else:
        # just checksum the contents of the first argument
        digest.update(filename)
    return digest.hexdigest()


def md5sum(filename,buffersize=16384):
    """Return md5 digest of file"""
    return cksm(filename,'md5',buffersize)


def sha1sum(filename,buffersize=16384):
    """Return sha1 digest of file"""
    return cksm(filename,'sha1',buffersize)


def sha256sum(filename,buffersize=16384):
    """Return sha256 digest of file"""
    return cksm(filename,'sha256',buffersize)


def sha512sum(filename,buffersize=16384):
    """Return sha512 digest of file"""
    return cksm(filename,'sha512',buffersize)


def load_cksm(sumfile, base_filename):
    """Load the checksum from a file"""
    for line in open(sumfile, 'r'):
        if os.path.basename(base_filename) in line:
            sum_cksm, name = line.strip('\n').split()
            return sum_cksm
    raise Exception('could not find checksum in file')


def check_cksm(file,type,sum):
    """Check a checksum of a file"""
    if not os.path.exists(file):
        return False
    # get checksum from file
    file_cksm = cksm(file,type)
    # load sum
    if os.path.isfile(sum):
        sum_cksm = load_cksm(sum, file)
    else:
        sum_cksm = sum
    # check sum
    logging.debug('file_cksm: %r', file_cksm)
    logging.debug('sum_cksm:  %r', sum_cksm)
    return (file_cksm == sum_cksm)


def check_md5sum(file,sum):
    """Check an md5sum of a file"""
    return check_cksm(file,'md5',sum)


def check_sha1sum(file,sum):
    """Check an sha1sum of a file"""
    return check_cksm(file,'sha1',sum)


def check_sha256sum(file,sum):
    """Check an sha256sum of a file"""
    return check_cksm(file,'sha256',sum)


def check_sha512sum(file,sum):
    """Check an sha512sum of a file"""
    return check_cksm(file,'sha512',sum)


# File and Directory Manipulation Functions #


def removedirs(path):
    try:
        if os.path.isdir(path):
            shutil.rmtree(path,True)
        else:
            os.remove(path)
    except Exception:
        pass


def copy(src,dest):
    parent_dir = os.path.dirname(dest)
    if not os.path.exists(parent_dir):
        logging.info('attempting to make parent dest dir %s',parent_dir)
        try:
            os.makedirs(parent_dir)
        except Exception:
            logging.error('failed to make dest directory for copy',exc_info=True)
            raise
    if os.path.isdir(src):
        logging.info('dircopy: %s to %s',src,dest)
        shutil.copytree(src,dest,symlinks=True)
    else:
        logging.info('filecopy: %s to %s',src,dest)
        shutil.copy2(src,dest)


# Network Functions #


def getInterfaces():
    """
    Get a list of available interfaces.

    Requires `psutil`.

    Returns:
        dict of {nic_name: {type: address}}
    """
    interfaces = {}

    ret = psutil.net_if_addrs()
    for nic_name in ret:
        n = {}
        for snic in ret[nic_name]:
            if not snic.address:
                continue
            if snic.family == socket.AF_INET:
                n['ipv4'] = snic.address
            elif snic.family == socket.AF_INET6:
                n['ipv6'] = snic.address
            elif snic.family == psutil.AF_LINK:
                n['mac'] = snic.address
        interfaces[nic_name] = n
    return interfaces


def get_local_ip_address():
    """Get the local (loopback) ip address"""
    try:
        return socket.gethostbyname('localhost')
    except Exception:
        return socket.gethostbyname(socket.getfqdn())


def gethostname():
    """Get hostname of this computer."""
    ret = socket.getfqdn()
    ret2 = socket.gethostname()
    if len(ret2) > len(ret):
        return ret2
    else:
        return ret


@contextmanager
def _http_helper(options={}, sync=True):
    """Set up an http session using requests"""
    if sync:
        session = Session
    else:
        session = AsyncSession
    with session(retries=10, backoff_factor=0.3) as s:
        if 'username' in options and 'password' in options:
            s.auth = (options['username'], options['password'])
        if 'sslcert' in options:
            if 'sslkey' in options:
                s.cert = (options['sslcert'], options['sslkey'])
            else:
                s.cert = options['sslcert']
        if 'cacert' in options:
            s.verify = options['cacert']
        if 'token' in options:
            s.headers.update({'Authorization': f'bearer {options["token"]}'})
        yield s


async def download(url, local, options={}):
    """Download a file, checksumming if possible"""
    local = os.path.expanduser(os.path.expandvars(local))
    url = os.path.expanduser(os.path.expandvars(url))
    if not isurl(url):
        if os.path.exists(url):
            url = 'file:'+url
        else:
            raise Exception("unsupported protocol %s" % url)

    # strip off query params
    if '?' in url:
        clean_url = url[:url.find('?')]
    elif '#' in url:
        clean_url = url[:url.find('#')]
    else:
        clean_url = url

    # fix local to be a filename to write to
    if local.startswith('file:'):
        local = local[5:]
    if os.path.isdir(local):
        local = os.path.join(local, os.path.basename(clean_url))

    logging.warning('wget(): src: %s, local: %s', url, local)

    # actually download the file
    try:
        if url.startswith('http'):
            logging.info('http from %s to %s', url, local)
            # http_proxy fix
            for k in os.environ:
                if k.lower() == 'http_proxy' and not os.environ[k].startswith('http'):
                    os.environ[k] = 'http://'+os.environ[k]

            def _d():
                with _http_helper(options) as s:
                    r = s.get(url, stream=True, timeout=300)
                    with open(local, 'wb') as f:
                        for chunk in r.iter_content(65536):
                            f.write(chunk)
                    r.raise_for_status()
            await asyncio.get_event_loop().run_in_executor(None, _d)
        elif url.startswith('file:'):
            url = url[5:]
            logging.info('copy from %s to %s', url, local)
            if os.path.exists(url):
                await asyncio.get_event_loop().run_in_executor(None, partial(copy, url, local))
        elif url.startswith('gsiftp:') or url.startswith('ftp:'):
            logging.info('gsiftp from %s to %s', url, local)
            await asyncio.get_event_loop().run_in_executor(None, partial(GridFTP.get, url, filename=local))
        else:
            raise Exception("unsupported protocol %s" % url)

        if not os.path.exists(local):
            raise Exception('download failed - file does not exist')
    except Exception:
        await asyncio.get_event_loop().run_in_executor(None, removedirs, local)
        raise

    return local


async def upload(local, url, checksum=True, options={}):  # noqa: C901
    """Upload a file, checksumming if possible"""
    local = os.path.expandvars(local)
    url = os.path.expandvars(url)
    if not isurl(url):
        if url.startswith('/'):
            url = 'file:'+url
        else:
            raise Exception("unsupported protocol %s" % url)
    if local.startswith('file:'):
        local = local[5:]
    if os.path.isdir(local):
        compress(local, 'tar')
        local += '.tar'

    logging.warning('wput(): local: %s, url: %s', local, url)

    if not os.path.exists(local):
        logging.warning('upload: local path, %s, does not exist', local)
        raise Exception('local file does not exist')

    chksum = sha512sum(local)
    chksum_type = 'sha512'
    if not checksum:
        logging.warning(f'not performing checksum {chksum_type}\n{checksum}: {url}')

    # actually upload the file
    if url.startswith('http'):
        logging.info('http from %s to %s', local, url)
        # http_proxy fix
        for k in os.environ:
            if k.lower() == 'http_proxy' and not os.environ[k].startswith('http'):
                os.environ[k] = 'http://'+os.environ[k]

        def _d():
            with _http_helper(options) as s:
                try:
                    with open(local, 'rb') as f:
                        r = s.put(url, timeout=300, data=f)
                    r.raise_for_status()
                except requests.exceptions.HTTPError as e:
                    if e.response.status_code != 405:
                        raise
                    else:
                        logging.warning('WebDav PUT not allowed, trying multipart upload')
                        with open(local, 'rb') as f:
                            m = MultipartEncoder(
                                fields={'field0': ('filename', f, 'text/plain')}
                            )
                            r = s.post(url, timeout=300, data=m,
                                       headers={'Content-Type': m.content_type})
                            r.raise_for_status()
                if checksum:  # get checksum
                    if 'ETAG' in r.headers:
                        md5 = r.headers['ETAG'].strip('"\'')
                        if md5sum(local) != md5:
                            raise Exception('http checksum error')
                    else:
                        r = s.get(url, stream=True, timeout=300)
                        try:
                            with open(local+'.tmp', 'wb') as f:
                                for chunk in r.iter_content(65536):
                                    f.write(chunk)
                            r.raise_for_status()
                            if sha512sum(local+'.tmp') != chksum:
                                raise Exception('http checksum error')
                        finally:
                            removedirs(local+'.tmp')
        await asyncio.get_event_loop().run_in_executor(None, _d)
    elif url.startswith('file:'):
        # use copy command
        url = url[5:]

        def _c():
            if os.path.exists(url):
                logging.warning('put: file already exists. overwriting!')
                removedirs(url)
            copy(local, url)
            if checksum and sha512sum(url) != chksum:
                raise Exception('file checksum error')
        await asyncio.get_event_loop().run_in_executor(None, _c)
    elif url.startswith('gsiftp:') or url.startswith('ftp:'):
        def _g():
            try:
                GridFTP.put(url, filename=local)
            except Exception:
                # because d-cache doesn't allow overwriting, try deletion
                GridFTP.delete(url)
                GridFTP.put(url, filename=local)
            if checksum and GridFTP.sha512sum(url) != chksum:
                raise Exception('gridftp checksum error')
        await asyncio.get_event_loop().run_in_executor(None, _g)
    else:
        raise Exception("unsupported protocol %s" % url)


def delete(url, options={}):
    """Delete a url or file"""
    url = os.path.expandvars(url)
    if (not isurl(url)) and os.path.exists(url):
        url = 'file:'+url

    if url.startswith('http'):
        logging.info('delete http: %s', url)
        with _http_helper(options) as s:
            r = s.delete(url, timeout=300)
            r.raise_for_status()
    elif url.startswith('file:'):
        url = url[5:]
        logging.info('delete file: %r', url)
        if os.path.exists(url):
            removedirs(url)
    elif url.startswith('gsiftp:') or url.startswith('ftp:'):
        logging.info('delete gsiftp: %r', url)
        GridFTP.rmtree(url)
    else:
        raise Exception("unsupported protocol %s" % url)


def isurl(url):
    """Determine if this is a supported protocol"""
    prefixes = ('file:','http:','https:','ftp:','ftps:','gsiftp:','pelican:','osdf:')
    try:
        return url.startswith(prefixes)
    except Exception:
        try:
            return reduce(lambda a,b: a or url.startswith(b), prefixes, False)
        except Exception:
            return False



import logging
import re

import pytest
import pytest_asyncio
from rest_tools.client import RestClient
from rest_tools.utils import Auth
import requests.exceptions

from iceprod.rest.auth import ROLES, GROUPS
from iceprod.services.server import Server


@pytest_asyncio.fixture
async def server(monkeypatch, port, mongo_url, mongo_clear):
    monkeypatch.setenv('CI_TESTING', '1')
    monkeypatch.setenv('PORT', str(port))
    monkeypatch.setenv('ICEPROD_API_ADDRESS', 'http://test.iceprod')
    monkeypatch.setenv('ICEPROD_CRED_ADDRESS', 'http://test.iceprod')

    s = Server()
    await s.start()

    auth = Auth('secret')

    def _add_to_data(data, attrs):
        logging.debug('attrs: %r', attrs)
        for item in attrs:
            key,value = item.split('=',1)
            d = data
            while '.' in key:
                k,key = key.split('.',1)
                if k not in d:
                    d[k] = {}
                d = d[k]
            if key not in d:
                d[key] = []
            d[key].append(value)

    def client(timeout=1, username='user', roles=[], groups=[], exp=10):
        data = {'preferred_username': username}
        for r in roles:
            _add_to_data(data, ROLES[r])
        for g in groups:
            _add_to_data(data, GROUPS[g])
        token = auth.create_token('username', expiration=exp, payload=data)
        return RestClient(f'http://localhost:{port}', token, timeout=timeout, retries=0)

    try:
        yield client
    finally:
        await s.stop()



import dataclasses

from wipac_dev_tools import from_environment_as_dataclass


@dataclasses.dataclass
class DefaultConfig:
    HOST: str = 'localhost'
    PORT: int = 8080
    DEBUG: bool = False
    OPENID_URL: str = ''
    OPENID_AUDIENCE: str = ''
    ICEPROD_API_ADDRESS: str = 'https://api.iceprod.icecube.aq'
    ICEPROD_API_CLIENT_ID: str = ''
    ICEPROD_API_CLIENT_SECRET: str = ''
    ICEPROD_CRED_ADDRESS: str = 'https://credentials.iceprod.icecube.aq'
    ICEPROD_CRED_CLIENT_ID: str = ''
    ICEPROD_CRED_CLIENT_SECRET: str = ''
    SERVICE_TIMEOUT: float = 900.
    SERVICE_SLEEP_SECS: float = 5.
    DB_URL: str = 'mongodb://localhost/iceprod'
    DB_TIMEOUT: int = 60
    DB_WRITE_CONCERN: int = 1
    PROMETHEUS_PORT: int = 0
    CI_TESTING: str = ''


def get_config() -> DefaultConfig:
    return from_environment_as_dataclass(DefaultConfig)



import asyncio
import dataclasses
import logging

from wipac_dev_tools import from_environment_as_dataclass

from .server import Server

# handle logging
setlevel = {
    'CRITICAL': logging.CRITICAL,  # execution cannot continue
    'FATAL': logging.CRITICAL,
    'ERROR': logging.ERROR,  # something is wrong, but try to continue
    'WARNING': logging.WARNING,  # non-ideal behavior, important event
    'WARN': logging.WARNING,
    'INFO': logging.INFO,  # initial debug information
    'DEBUG': logging.DEBUG  # the things no one wants to see
}


@dataclasses.dataclass(frozen=True)
class DefaultConfig:
    LOG_LEVEL: str = 'INFO'


config = from_environment_as_dataclass(DefaultConfig)

if config.LOG_LEVEL.upper() not in setlevel:
    raise Exception('LOG_LEVEL is not a proper log level')
logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'

logging.basicConfig(format=logformat, level=setlevel[config.LOG_LEVEL.upper()])


# start server
async def main():
    s = Server()
    await s.start()
    try:
        await asyncio.Event().wait()
    finally:
        await s.stop()


asyncio.run(main())



from math import isclose
import time
from urllib.parse import urlparse, parse_qs

from iceprod.s3 import S3


async def test_create_bucket(s3conn):
    s = S3(address='http://localhost:5000', access_key='XXXX', secret_key='XXXX', bucket='test', mock_s3=s3conn)
    await s.create_bucket()

    assert any(b['Name'] == 'test' for b in s3conn.list_buckets()['Buckets'])


async def test_s3_get(s3conn):
    s = S3(address='http://localhost:5000', access_key='XXXX', secret_key='XXXX', bucket='test', mock_s3=s3conn)
    await s.create_bucket()

    r = await s.get('foo')
    assert not r

async def test_s3_put(s3conn):
    s = S3(address='http://localhost:5000', access_key='XXXX', secret_key='XXXX', bucket='test', mock_s3=s3conn)
    await s.create_bucket()

    data = 'some data'
    await s.put('foo', data)
    r = await s.get('foo')
    assert r == data

async def test_s3_get_presigned(s3conn):
    s = S3(address='http://localhost:5000', access_key='XXXX', secret_key='XXXX', bucket='test', mock_s3=s3conn)
    await s.create_bucket()

    data = 'some data'
    await s.put('foo', data)
    exp = 1000
    r = s.get_presigned('foo', expiration=exp)
    o = urlparse(r)
    assert o.path.endswith('/foo')
    q = parse_qs(o.query)
    assert 'Expires' in q
    assert isclose(float(q['Expires'][0]), time.time()+exp, abs_tol=1.)

async def test_s3_put_presigned(s3conn):
    s = S3(address='http://localhost:5000', access_key='XXXX', secret_key='XXXX', bucket='test', mock_s3=s3conn)
    await s.create_bucket()

    exp = 1000
    r = s.put_presigned('foo', expiration=exp)
    o = urlparse(r)
    assert o.path.endswith('/foo')
    q = parse_qs(o.query)
    assert 'Expires' in q
    assert isclose(float(q['Expires'][0]), time.time()+exp, abs_tol=1.)

async def test_s3_exists(s3conn):
    s = S3(address='http://localhost:5000', access_key='XXXX', secret_key='XXXX', bucket='test', mock_s3=s3conn)
    await s.create_bucket()

    r = await s.exists('foo')
    assert r is False

    await s.put('foo', 'data')
    r = await s.exists('foo')
    assert r is True

async def test_s3_delete(s3conn):
    s = S3(address='http://localhost:5000', access_key='XXXX', secret_key='XXXX', bucket='test', mock_s3=s3conn)
    await s.create_bucket()

    await s.put('foo', 'data')
    r = await s.exists('foo')
    assert r is True

    await s.delete('foo')
    r = await s.exists('foo')
    assert r is False

async def test_s3_list(s3conn):
    s = S3(address='http://localhost:5000', access_key='XXXX', secret_key='XXXX', bucket='test', mock_s3=s3conn)
    await s.create_bucket()

    await s.put('foo', 'data')
    await s.put('bar/1', 'data')
    await s.put('bar/2', 'data2')

    r = await s.list()
    assert r == {
        'foo': 4,
        'bar': {
            '1': 4,
            '2': 5,
        }
    }

    r = await s.list(prefix='bar', recursive=False)
    assert r == {
        'bar/1': 4,
        'bar/2': 5,
    }

async def test_s3_rmtree(s3conn):
    s = S3(address='http://localhost:5000', access_key='XXXX', secret_key='XXXX', bucket='test', mock_s3=s3conn)
    await s.create_bucket()

    await s.put('foo', 'data')
    await s.put('bar/1', 'data')
    await s.put('bar/2', 'data2')

    await s.rmtree('bar')

    r = await s.list(recursive=False)
    assert r == {
        'foo': 4,
    }



import pytest
import requests.exceptions

import iceprod.server.states


async def test_rest_jobs_add(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    ret = await client.request('GET', f'/jobs/{job_id}')
    for k in data:
        assert k in ret
        assert data[k] == ret[k]
    for k in ('status','status_changed'):
        assert k in ret
    assert ret['status'] == 'processing'


async def test_rest_jobs_add_bad_role(server):
    client = server(roles=['user'], groups=['users'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', '/jobs', data)
    assert exc_info.value.response.status_code == 403


async def test_rest_jobs_patch(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    new_data = {
        'status': 'suspended',
    }
    ret = await client.request('PATCH', f'/jobs/{job_id}', new_data)

    ret = await client.request('GET', f'/jobs/{job_id}')
    assert ret['status'] == 'suspended'


async def test_rest_jobs_dataset_get(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/jobs')
    assert job_id in ret
    for k in data:
        assert k in ret[job_id]
        assert data[k] == ret[job_id][k]


async def test_rest_jobs_dataset_get_details(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/jobs/{job_id}')
    for k in data:
        assert k in ret
        assert data[k] == ret[k]
    for k in ('status','status_changed'):
        assert k in ret
    assert ret['status'] == 'processing'


async def test_rest_jobs_empty(server):
    client = server(roles=['system'])

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('GET', '/datasets/foo/jobs/bar')
    assert exc_info.value.response.status_code == 404


async def test_rest_jobs_dataset_set_status(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    new_data = {'status': 'errors'}
    ret = await client.request('PUT', f'/datasets/{data["dataset_id"]}/jobs/{job_id}/status', new_data)

    ret = await client.request('GET', f'/jobs/{job_id}')
    assert ret['status'] == new_data['status']


async def test_rest_jobs_dataset_bulk_set_status(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    new_data = {'jobs': [job_id]}
    ret = await client.request('POST', f'/datasets/{data["dataset_id"]}/job_actions/bulk_status/errors', new_data)

    ret = await client.request('GET', f'/jobs/{job_id}')
    assert ret['status'] == 'errors'


async def test_rest_jobs_dataset_bulk_suspend(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    new_data = {}
    ret = await client.request('POST', f'/datasets/{data["dataset_id"]}/job_actions/bulk_suspend', new_data)

    ret = await client.request('GET', f'/jobs/{job_id}')
    assert ret['status'] == 'suspended'


async def test_rest_jobs_dataset_bulk_suspend_specific(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    data = {
        'dataset_id': 'foo',
        'job_index': 1,
    }
    ret = await client.request('POST', '/jobs', data)
    job_id2 = ret['result']

    new_data = {'jobs': [job_id]}
    ret = await client.request('POST', f'/datasets/{data["dataset_id"]}/job_actions/bulk_suspend', new_data)

    ret = await client.request('GET', f'/jobs/{job_id}')
    assert ret['status'] == 'suspended'
    ret = await client.request('GET', f'/jobs/{job_id2}')
    assert ret['status'] == iceprod.server.states.JOB_STATUS_START


async def test_rest_jobs_dataset_bulk_reset(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
        'status': 'errors',
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    new_data = {}
    ret = await client.request('POST', f'/datasets/{data["dataset_id"]}/job_actions/bulk_reset', new_data)

    ret = await client.request('GET', f'/jobs/{job_id}')
    assert ret['status'] == iceprod.server.states.JOB_STATUS_START


async def test_rest_jobs_dataset_bulk_reset_specific(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
        'status': 'errors',
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    data = {
        'dataset_id': 'foo',
        'job_index': 1,
        'status': 'errors',
    }
    ret = await client.request('POST', '/jobs', data)
    job_id2 = ret['result']

    new_data = {'jobs': [job_id]}
    ret = await client.request('POST', f'/datasets/{data["dataset_id"]}/job_actions/bulk_reset', new_data)

    ret = await client.request('GET', f'/jobs/{job_id}')
    assert ret['status'] == iceprod.server.states.JOB_STATUS_START
    ret = await client.request('GET', f'/jobs/{job_id2}')
    assert ret['status'] == 'errors'


async def test_rest_jobs_dataset_bulk_reset_blocked(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
        'status': 'complete',
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    with pytest.raises(Exception):
        await client.request('POST', f'/datasets/{data["dataset_id"]}/job_actions/bulk_reset', new_data)

    ret = await client.request('GET', f'/jobs/{job_id}')
    assert ret['status'] == 'complete'


async def test_rest_jobs_dataset_bulk_hardreset(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
        'status': 'complete',
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    new_data = {}
    ret = await client.request('POST', f'/datasets/{data["dataset_id"]}/job_actions/bulk_hard_reset', new_data)

    ret = await client.request('GET', f'/jobs/{job_id}')
    assert ret['status'] == iceprod.server.states.JOB_STATUS_START


async def test_rest_jobs_dataset_bulk__hardreset_specific(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
        'status': 'complete',
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    data = {
        'dataset_id': 'foo',
        'job_index': 1,
        'status': 'complete',
    }
    ret = await client.request('POST', '/jobs', data)
    job_id2 = ret['result']

    new_data = {'jobs': [job_id]}
    ret = await client.request('POST', f'/datasets/{data["dataset_id"]}/job_actions/bulk_hard_reset', new_data)

    ret = await client.request('GET', f'/jobs/{job_id}')
    assert ret['status'] == iceprod.server.states.JOB_STATUS_START
    ret = await client.request('GET', f'/jobs/{job_id2}')
    assert ret['status'] == 'complete'


async def test_rest_jobs_dataset_summaries_status(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/job_summaries/status')
    assert ret == {'processing': [job_id]}
    

async def test_rest_jobs_dataset_counts_status(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_index': 0,
    }
    ret = await client.request('POST', '/jobs', data)
    job_id = ret['result']

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/job_counts/status')
    assert ret == {'processing': 1}



#!/usr/bin/env python

import os
import subprocess
import argparse
import json
import glob
import logging
import asyncio

from rest_tools.client import RestClient
from iceprod.server.scheduled_tasks.buffer_jobs_tasks import run

logger = logging.getLogger()


def main():
    parser = argparse.ArgumentParser(description='clean up processing tasks not in pilots')
    parser.add_argument('-t', '--token', help='auth token')
    parser.add_argument('-d','--dataset', help='dataset id')
    parser.add_argument('-n','--num', type=int, default=20000,
                        help='max number of jobs to buffer')
    parser.add_argument('--debug', action='store_true', help='debug')
    
    args = parser.parse_args()
    args = vars(args)

    logging.basicConfig(level=logging.DEBUG if args['debug'] else logging.INFO)
    
    rpc = RestClient('https://materialization.iceprod.icecube.aq', args['token'])

    asyncio.run(run(rpc, only_dataset=args['dataset'], num=args['num'], run_once=True, debug=args['debug']))

if __name__ == '__main__':
    main()




import pytest
import requests
from unittest.mock import AsyncMock, MagicMock

from iceprod.common.mongo_queue import AsyncMongoQueue
from iceprod.services.config import get_config
from iceprod.services.service import main
import iceprod.services.actions.submit

async def test_submit(monkeypatch, mongo_url, mongo_clear):
    monkeypatch.setenv('CI_TESTING', '1')
    action_mock = MagicMock()
    action_mock.return_value.run = AsyncMock(return_value=None)
    monkeypatch.setattr(iceprod.services.actions.submit, 'Action', action_mock)

    config = get_config()
    message_queue = AsyncMongoQueue(
        url=config.DB_URL,
        collection_name='services_queue',
        timeout=config.DB_TIMEOUT,
        write_concern=config.DB_WRITE_CONCERN
    )

    await main()
    assert action_mock.return_value.run.called == False

    await message_queue.push({
        'type': 'submit',
        'foo': 'bar'
    })

    await main()

    assert action_mock.return_value.run.called
    assert action_mock.return_value.run.call_args.args[0].payload == {'foo': 'bar'}

    await message_queue.close()


import dataclasses

from wipac_dev_tools import from_environment_as_dataclass


@dataclasses.dataclass
class DefaultConfig:
    HOST : str = 'localhost'
    PORT : int = 8080
    DEBUG : bool = False
    OPENID_URL : str = ''
    OPENID_AUDIENCE : str = ''
    ICEPROD_CRED_ADDRESS : str = 'https://credentials.iceprod.icecube.aq'
    ICEPROD_CRED_CLIENT_ID : str = ''
    ICEPROD_CRED_CLIENT_SECRET : str = ''
    DB_URL: str = 'mongodb://localhost/iceprod'
    DB_TIMEOUT: int = 60
    DB_WRITE_CONCERN: int = 1
    PROMETHEUS_PORT: int = 0
    S3_ADDRESS: str = ''
    S3_ACCESS_KEY: str = ''
    S3_SECRET_KEY: str = ''
    MAX_BODY_SIZE: int = 10**9
    ROUTE_STATS_WINDOW_SIZE: int = 1000
    ROUTE_STATS_WINDOW_TIME: int = 3600
    ROUTE_STATS_TIMEOUT: int = 60
    CI_TESTING: str = ''


def get_config() -> DefaultConfig:
    return from_environment_as_dataclass(DefaultConfig, obfuscate_log_vars=['DB_URL'])



from __future__ import absolute_import, division, print_function


import os
import sys
import time
import signal
import argparse
import subprocess
import logging
import importlib
import json
import sqlite3
import tempfile
import shutil
import glob

import requests
import psutil

# add iceprod to PYTHONPATH
curdir = os.getcwd()
integration_dir = os.path.dirname(os.path.abspath(__file__))
iceprod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0,curdir)
sys.path.insert(0,iceprod_dir)
sys.path.insert(0,integration_dir)
if 'PYTHONPATH' in os.environ:
    os.environ['PYTHONPATH'] = '{}:{}:{}:{}'.format(integration_dir,iceprod_dir,curdir,os.environ['PYTHONPATH'])
else:
    os.environ['PYTHONPATH'] = '{}:{}:{}'.format(integration_dir,iceprod_dir,curdir)

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger('integration_tests')

from iceprod.core.jsonRPCclient import JSONRPC


parser = argparse.ArgumentParser()
parser.add_argument('-p','--port', type=int, default=37284, help='iceprod port')
parser.add_argument('--pilots', action='store_true', help='submit pilot jobs')
parser.add_argument('--timeout', type=int, default=3600, help='test timeout')
parser.add_argument('datasets', action='append', nargs='?')
args = parser.parse_args()
if args.datasets and args.datasets[0] is None:
    args.datasets = []

tmpdir = tempfile.mkdtemp(dir=curdir)
os.chdir(tmpdir)
os.environ['I3PROD'] = tmpdir
def cleanup():
    procs = psutil.Process().children(recursive=True)
    for p in procs:
        p.terminate()
    gone, alive = psutil.wait_procs(procs, timeout=1)
    for p in alive:
        p.kill()

    try:
        subprocess.call(['condor_rm','-all'])
    except Exception:
        pass
    os.chdir(curdir)
    if os.path.exists(tmpdir):
        shutil.rmtree(tmpdir)

# handle any signals
def handler1(signum, frame):
    logger.warn('Signal handler called with signal %s' % signum)
    logger.warn('Exiting...')
    cleanup()
    sys.exit(1)
signal.signal(signal.SIGQUIT, handler1)
signal.signal(signal.SIGINT, handler1)

# start testing
logger.info('starting...')

site_temp = os.path.join(tmpdir,'site_temp')
os.mkdir(site_temp)

# server config
port = args.port
cfg = {
    "modules":{
        "master_updater":False,
    },
    "logging":{
        "level":"DEBUG"
    },
    "schedule":{
        "buffer_jobs_tasks":"every 1 minutes",
    },
    "queue":{
        "a":{
            "type":"condor",
            "description":"test",
            "tasks_on_queue":[30,50,20],
            "pilots_on_queue":[30,50,20],
            "software_dir":os.environ['ICEPRODROOT'],
            "iceprod_dir":iceprod_dir
        },
        "queue_interval":30,
        "submit_pilots":args.pilots,
        "submit_dir":os.path.join(tmpdir,'submit'),
        "site_temp":site_temp,
        "debug":True,
    },
    "system":{
        "ssl":False,
    },
    "webserver":{
        "port":port,
        "tornado_port":port+1,
    },
}
if not os.path.exists('etc'):
    os.mkdir('etc')
with open('etc/iceprod_config.json','w') as f:
    json.dump(cfg,f)

# start iceprod server instance
start_time = time.time()
iceprod_server = subprocess.Popen([os.path.join(iceprod_dir,'bin/iceprod_server.py'),'-n'],cwd=tmpdir)
time.sleep(5)
if iceprod_server.poll() is not None:
    cleanup()
    raise Exception('server died unexpectedly')

try:
    # add passkey
    with sqlite3.connect('db') as conn:
        sql = 'insert into passkey (passkey_id,auth_key,expire,user_id) values '
        sql += '("blah","passkey","3000-01-01T00:00:00","")'
        conn.execute(sql)

    client = JSONRPC('http://localhost:%d/jsonrpc'%port,passkey='passkey')
    def submit_dataset(cfg):
        desc = None
        if 'description' in cfg:
            desc = cfg['description']
        return client.submit_dataset(cfg, njobs=10, description=desc)
    def wait_for_dataset(dataset_id):
        logger.info('waiting on dataset %s',dataset_id)
        while True:
            tasks = {'complete':0,'failed':0,'suspended':0}
            tasks.update(client.public_get_number_of_tasks_in_each_state(dataset_id))
            if tasks['complete'] == sum(tasks.values()) and tasks['complete'] > 10:
                return
            if tasks['failed'] | tasks['suspended'] > 1:
                raise Exception('dataset failed')
            time.sleep(60)
            if time.time()-start_time > args.timeout:
                raise Exception('over timeout limit')

    # submit datasets
    dataset_ids = []
    files = glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)),'*.json'))
    for dataset in files:
        if args.datasets and not any(x in dataset for x in args.datasets):
            continue
        logger.info('starting dataset %s', os.path.basename(dataset))
        cfg = json.load(open(dataset))
        dataset_ids.append(submit_dataset(cfg))

    # wait for successful completion of datasets
    for d in dataset_ids:
        wait_for_dataset(d)
finally:
    cleanup()

logger.info('success!')



"""
Test script for scheduled_tasks/queue_tasks
"""

import logging
from unittest.mock import MagicMock

import pytest
from iceprod.scheduled_tasks import queue_tasks

logger = logging.getLogger('scheduled_tasks_queue_tasks_test')


async def test_200_run():
    config = queue_tasks.default_config.copy()
    rc = MagicMock()
    async def client(method, url, args=None):
        if url == '/datasets/foo':
            return {'priority': 2}
        elif url == '/task_counts/status':
            return {'idle': 100, 'waiting': 2}
        elif url == '/tasks':
            return {'tasks': [{'task_id': 'task1'}]}
        elif url == '/task_actions/waiting' and method == 'POST':
            client.called = True
            return {'waiting': 98}
        else:
            raise Exception()
    client.called = False
    rc.request = client
    await queue_tasks.run(rc, config, debug=True)
    assert client.called
    
    async def client(method, url, args=None):
        if url == '/task_counts/status':
            return {}
        elif url == '/task_actions/waiting' and method == 'POST':
            client.called = True
            return {'waiting': 0}
        else:
            raise Exception()
    client.called = False
    rc.request = client
    await queue_tasks.run(rc, config, debug=True)
    assert not client.called
    
    async def client(method, url, args=None):
        if url == '/datasets/foo':
            return {'priority': 2}
        elif url.startswith('/task_counts/status'):
            return {}
        elif url == '/task_actions/waiting' and method == 'POST':
            client.called = True
            return {'waiting': 0}
        else:
            raise Exception()
    client.called = False
    rc.request = client
    await queue_tasks.run(rc, config, debug=True)
    assert not client.called

    async def client(method, url, args=None):
        if url == '/datasets/foo':
            return {'priority': 2}
        elif url.startswith('/task_counts/status'):
            return {'idle': 100,'waiting': 100000}
        elif url == '/tasks':
            return {'tasks': []}
        elif url == '/task_actions/waiting' and method == 'POST':
            client.called = True
            return {'waiting': 0}
        else:
            raise Exception()
    client.called = False
    rc.request = client
    await queue_tasks.run(rc, config, debug=True)
    assert not client.called


async def test_300_run():
    config = queue_tasks.default_config.copy()
    rc = MagicMock()
    async def client(method, url, args=None):
        if url.startswith('/task_counts/status'):
            client.called = True
            return {'idle': 100, 'waiting': 100000}
        else:
            raise Exception()
    client.called = False
    rc.request = client
    with pytest.raises(Exception):
        await queue_tasks.run(rc, config, debug=True)
    assert client.called

    # internally catch the error
    await queue_tasks.run(rc, config)



#!/usr/bin/env python3

import os
from pathlib import Path
import sys


"""
This is a PreCmd script designed to rewrite input file names.
"""

for arg in sys.argv[1:]:
    try:
        name, newname = arg.split('=',1)
    except Exception:
        print('Bad formatting, cannot parse')
        raise
    try:
        if not os.path.exists(name):
            continue
        if '/' in newname:
            Path(newname).parent.mkdir(parents=True, exist_ok=True)
        os.rename(name, newname)
    except Exception:
        print(f'Cannot move inputfile {name} to {newname}', file=sys.stderr)
        raise



import logging

from .base import authenticated, PublicHandler

logger = logging.getLogger('website-job')


class JobBrowse(PublicHandler):
    """Handle /job urls"""
    @authenticated
    async def get(self, dataset_id):
        assert self.rest_client
        status = self.get_argument('status',default=None)
        passkey = self.auth_access_token

        jobs = await self.rest_client.request('GET', '/datasets/{}/jobs'.format(dataset_id))
        if status:
            for t in list(jobs):
                if jobs[t]['status'] != status:
                    del jobs[t]
                    continue
        self.render('job_browse.html', jobs=jobs, passkey=passkey)


class Job(PublicHandler):
    """Handle /job urls"""
    @authenticated
    async def get(self, dataset_id, job_id):
        assert self.rest_client
        status = self.get_argument('status',default=None)
        passkey = self.auth_access_token

        dataset = await self.rest_client.request('GET', '/datasets/{}'.format(dataset_id))
        job = await self.rest_client.request('GET', '/datasets/{}/jobs/{}'.format(dataset_id,job_id))
        args = {'job_id': job_id}
        if status:
            args['status'] = status
        tasks = await self.rest_client.request('GET', f'/datasets/{dataset_id}/tasks', args)
        job['tasks'] = list(tasks.values())
        job['tasks'].sort(key=lambda x:x['task_index'])
        self.render('job_detail.html', dataset=dataset, job=job, passkey=passkey)



from dataclasses import asdict, dataclass
import json
from typing import Any

from tornado.web import HTTPError

from iceprod.common.mongo_queue import Message
from iceprod.rest.auth import attr_auth, authorization
from iceprod.services.base import AuthData, BaseAction, BaseHandler, HandlerTypes, TimeoutException
from .materialize import Materialize


class DatasetHandler(BaseHandler):
    @authorization(roles=['admin', 'system', 'user'])
    @attr_auth(arg='dataset_id', role='read')
    async def get(self, dataset_id: str):
        assert self.action
        type_ = self.action.action_type
        ret = await self.message_queue.lookup_by_payload({'type': type_, 'dataset_id': dataset_id})
        if not ret:
            self.send_error(404, reason="Request not found")
        else:
            self.write({
                'id': ret.uuid,
                'status': ret.status,
            })

    @authorization(roles=['admin', 'system', 'user'])
    @attr_auth(arg='dataset_id', role='write')
    async def post(self, dataset_id: str):
        assert self.action

        args = json.loads(self.request.body) if self.request.body else {}
        args['dataset_id'] = dataset_id

        auth = AuthData(
            username=self.current_user,
            groups=self.auth_groups,
            roles=self.auth_roles,
            token=self.auth_data,
        )

        id_ = await self.action.create(args, auth_data=auth)

        self.set_status(201)
        self.write({'result': id_})


@dataclass
class Fields:
    dataset_id: str
    set_status: str = 'idle'
    num: int = 1000


class Action(BaseAction):
    PRIORITY = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._materialize = Materialize(rest_client=self._api_client)

    def extra_handlers(self) -> HandlerTypes:
        """Return handlers"""
        return [
            (r'/datasets/(?P<dataset_id>\w+)', DatasetHandler, {}),
        ]

    async def create(self, args: dict[str, Any], *, auth_data: AuthData) -> str:
        """
        Validates a new materialization request.

        Deduplicates with existing requests with the same dataset_id.

        Arguments:
            args: dict of args for request

        Returns:
            str: request id
        """
        # validate request
        try:
            data = Fields(**args)
            if isinstance(data.num, str):
                data.num = int(data.num)
        except Exception as e:
            raise HTTPError(400, reason=str(e))

        # deduplicate on dataset_id
        return await self._push(payload=asdict(data), filter_payload={'dataset_id': data.dataset_id}, priority=self.PRIORITY)

    async def run(self, message: Message) -> None:
        """Run materialization"""
        data = message.payload
        if not self._materialize:
            return None
        self._logger.info(f'running materialization request: {data}')
        kwargs = {}
        if 'dataset_id' in data and data['dataset_id']:
            kwargs['only_dataset'] = data['dataset_id']
        if 'num' in data and data['num']:
            kwargs['num'] = data['num']
        if 'set_status' in data and data['set_status']:
            kwargs['set_status'] = data['set_status']
        ret = await self._materialize.run_once(**kwargs)
        self._logger.info('ret: %r', ret)

        if not ret:
            self._logger.warning('materialization request took too long, bumping to end of queue for another run')
            raise TimeoutException()
        return None



from dataclasses import asdict
import re
from datetime import datetime
import logging
from typing import Any
from unittest.mock import AsyncMock
from uuid import uuid4

import pytest

from iceprod.common.mongo_queue import Message
from iceprod.services.actions.dataset_status import Action, Fields


@pytest.mark.parametrize('args,error', [
    ({
        'dataset_id': 'd1',
        'action': 'reset',
    }, False),
    ({
        'dataset_id': 'd1',
        'action': 'foo',
    }, 'invalid action'),
])
async def test_dataset_status_server_request(args, error, server, requests_mock):
    requests_mock.register_uri('GET', re.compile('localhost'), real_http=True)
    requests_mock.register_uri('POST', re.compile('localhost'), real_http=True)

    requests_mock.post('http://test.iceprod/auths', status_code=200, json={})

    client = server(roles=['user'], groups=['users'])

    if error:
        with pytest.raises(Exception, match=error):
            await client.request('POST', '/actions/dataset_status', args)
    else:
        ret = await client.request('POST', '/actions/dataset_status', args)
        assert 'result' in ret

        ret = await client.request('GET', f'/actions/dataset_status/{ret["result"]}')
        assert ret['status'] == 'queued'
        for k in args:
            assert ret['payload'][k] == args[k]
        assert ret['payload']['progress'] == 0


async def test_dataset_status_server_request_bad_auth(server, requests_mock):
    requests_mock.register_uri('GET', re.compile('localhost'), real_http=True)
    requests_mock.register_uri('POST', re.compile('localhost'), real_http=True)

    requests_mock.post('http://test.iceprod/auths', status_code=403, json={})

    client = server(roles=['user'], groups=['users'])
    args = {
        'dataset_id': 'd1',
        'action': 'reset',
    }
    with pytest.raises(Exception, match='auth failed'):
        await client.request('POST', '/actions/dataset_status', args)


async def test_dataset_status_server_request_no_auth(server, requests_mock):
    requests_mock.register_uri('GET', re.compile('localhost'), real_http=True)
    requests_mock.register_uri('POST', re.compile('localhost'), real_http=True)

    requests_mock.post('http://test.iceprod/auths', status_code=403, json={})

    client = server(roles=['system'], groups=[])
    args = {
        'dataset_id': 'd1',
        'action': 'reset',
    }
    await client.request('POST', '/actions/dataset_status', args)


@pytest.mark.parametrize('args', [
    {
        'dataset_id': 'd1',
        'action': 'reset',
    },
])
async def test_dataset_status_run(args):
    queue = AsyncMock()
    api_client = AsyncMock()
    cred_client = AsyncMock()
    submit = Action(queue, logging.getLogger('action'), api_client, cred_client)

    message = Message(
        payload=asdict(Fields(**args)),
        uuid=uuid4().hex,
        status='queued',
        priority=0,
        created_at=datetime.now(),
    )

    async def api(method: str, path: str, args: dict[str, Any] = {}):
        if method == 'POST' and path == f'/datasets/{message.payload["dataset_id"]}/task_actions/bulk_{message.payload["action"]}':
            return {}
        elif method == 'POST' and path == f'/datasets/{message.payload["dataset_id"]}/job_actions/bulk_{message.payload["action"]}':
            return {}
        elif method == 'PUT' and path == f'/datasets/{message.payload["dataset_id"]}/status':
            return {}
        elif method == 'GET' and path == f'/datasets/{message.payload["dataset_id"]}':
            return {
                'status': 'processing'
            }
        elif method == 'GET' and path == f'/datasets/{message.payload["dataset_id"]}/jobs':
            return {
                'j1': {'job_id': 'j1'},
                'j2': {'job_id': 'j2'},
            }
        else:
            raise Exception('invalid path')

    api_client.request.side_effect = api

    await submit.run(message)

    assert queue.update_payload.call_args[0][0] == message.uuid
    assert queue.update_payload.call_args[0][1] == {'progress': 100}



"""
Server for iceprod background services.
"""
import json
import logging
from pathlib import Path
import pkgutil
from typing import Any, Self

from prometheus_client import Info, start_http_server
from rest_tools.client import RestClient, ClientCredentialsAuth
from rest_tools.server import RestServer
from tornado.web import HTTPError
from tornado.web import RequestHandler as TornadoRequestHandler

from .config import get_config
from .base import AuthData, BaseAction, BaseHandler
from iceprod.common.mongo_queue import AsyncMongoQueue
from iceprod.util import VERSION_STRING
from iceprod.common.prom_utils import AsyncMonitor
from iceprod.rest.auth import authorization
from iceprod.rest.base_handler import IceProdRestConfig
from iceprod.server.util import nowstr

logger = logging.getLogger('server')


class RequestHandler(BaseHandler):
    """
    Handle basic requests.
    """
    @authorization(roles=['admin', 'system', 'user'])
    async def get(self):
        """
        Get a count of request statuses.

        Returns:
            dict: {status: count}
        """
        assert self.action
        type_ = self.action.action_type
        counts = {}
        for status in ('queued', 'processing', 'error', 'complete'):
            ret = await self.message_queue.count({'payload.type': type_, 'status': status})
            counts[status] = ret
        self.write(counts)

    @authorization(roles=['admin', 'system', 'user'])
    async def post(self):
        """
        Create basic request.

        On success, returns http code 201.

        Returns:
            dict: {result: id}
        """
        assert self.action

        args = json.loads(self.request.body) if self.request.body else {}

        auth = AuthData(
            username=self.current_user,
            groups=self.auth_groups,
            roles=self.auth_roles,
            token=self.auth_data,
        )

        id_ = await self.action.create(args, auth_data=auth)

        # return success
        self.set_status(201)
        self.write({'result': id_})


class StatusHandler(BaseHandler):
    """
    Handle status requests.
    """
    @authorization(roles=['admin', 'system', 'user'])
    async def get(self, id_):
        """
        Get the status of a request.

        If id is invalid, returns http code 404.

        Args:
            id_: request id

        Returns:
            dict: metadata
        """
        ret = await self.message_queue.get_status(id_)
        if not ret:
            self.send_error(404, reason="Request not found")
        else:
            payload = await self.message_queue.get_payload(id_)
            if ret == 'error':
                error_message = await self.message_queue.get_error(id_)
            else:
                error_message = ''
            self.write({
                'id': id_,
                'status': ret,
                'error_message': error_message,
                'payload': payload,
            })


class HealthHandler(BaseHandler):
    """
    Handle health requests.
    """
    async def get(self):
        """
        Get health status.

        Returns based on exit code, 200 = ok, 400 = failure
        """
        status = {
            'now': nowstr(),
            'num_requests': -1,
        }

        try:
            ret = await self.message_queue.count()
        except Exception:
            logger.info('bad db request', exc_info=True)
            self.send_error(500, reason='bad db request')
        else:
            if ret is None:
                self.send_error(500, reason='bad db result')
            else:
                status['num_requests'] = ret
                self.write(status)


class Error(TornadoRequestHandler):
    def prepare(self):
        raise HTTPError(404, 'invalid api route')


class Server:
    def __init__(self: Self):
        config = get_config()

        rest_config: dict[str, Any] = {
            'debug': config.DEBUG,
        }
        if config.OPENID_URL:
            logging.info(f'enabling auth via {config.OPENID_URL} for aud "{config.OPENID_AUDIENCE}"')
            rest_config['auth'] = {
                'openid_url': config.OPENID_URL,
                'audience': config.OPENID_AUDIENCE,
            }
        elif config.CI_TESTING:
            rest_config['auth'] = {
                'secret': 'secret',
            }
        else:
            raise RuntimeError('OPENID_URL not specified, and CI_TESTING not enabled!')

        # enable monitoring
        self.prometheus_port = config.PROMETHEUS_PORT if config.PROMETHEUS_PORT > 0 else None
        self.async_monitor = None

        self.message_queue = AsyncMongoQueue(
            url=config.DB_URL,
            collection_name='services_queue',
            extra_indexes={'type_index': {'keys': 'type', 'unique': False}},
            timeout=config.DB_TIMEOUT,
            write_concern=config.DB_WRITE_CONCERN
        )

        rest_client: RestClient
        if config.ICEPROD_API_CLIENT_ID and config.ICEPROD_API_CLIENT_SECRET:
            logging.info(f'enabling auth via {config.OPENID_URL} for aud "{config.OPENID_AUDIENCE}"')
            rest_client = ClientCredentialsAuth(
                address=config.ICEPROD_API_ADDRESS,
                token_url=config.OPENID_URL,
                client_id=config.ICEPROD_API_CLIENT_ID,
                client_secret=config.ICEPROD_API_CLIENT_SECRET,
            )
        elif config.CI_TESTING:
            rest_client = RestClient(config.ICEPROD_API_ADDRESS, timeout=1, retries=0)
        else:
            raise RuntimeError('ICEPROD_API_CLIENT_ID or ICEPROD_API_CLIENT_SECRET not specified, and CI_TESTING not enabled!')

        cred_client: RestClient
        if config.ICEPROD_CRED_CLIENT_ID and config.ICEPROD_CRED_CLIENT_SECRET:
            logging.info(f'enabling auth via {config.OPENID_URL} for aud "{config.OPENID_AUDIENCE}"')
            cred_client = ClientCredentialsAuth(
                address=config.ICEPROD_CRED_ADDRESS,
                token_url=config.OPENID_URL,
                client_id=config.ICEPROD_CRED_CLIENT_ID,
                client_secret=config.ICEPROD_CRED_CLIENT_SECRET,
            )
        elif config.CI_TESTING:
            cred_client = RestClient(config.ICEPROD_CRED_ADDRESS, timeout=1, retries=0)
        else:
            raise RuntimeError('ICEPROD_CRED_CLIENT_ID or ICEPROD_CRED_CLIENT_SECRET not specified, and CI_TESTING not enabled!')

        handler_args = IceProdRestConfig(rest_config)
        handler_args['message_queue'] = self.message_queue
        handler_args['rest_client'] = rest_client

        server = RestServer(debug=config.DEBUG)

        # find all actions
        plugin_path = str(Path(__file__).parent / 'actions')
        logger.info('action path %s', plugin_path)
        for _, name, _ in pkgutil.iter_modules([plugin_path]):
            logger.info('configuring action %s', name)
            action_class = pkgutil.resolve_name(f'iceprod.services.actions.{name}:Action')
            action : BaseAction = action_class(queue=self.message_queue, logger=logger, api_client=rest_client, cred_client=cred_client)

            handler_args2 = handler_args.copy()
            handler_args2['action'] = action
            server.add_route(f'/actions/{name}', RequestHandler, handler_args2)
            server.add_route(rf'/actions/{name}/(?P<id_>\w+)', StatusHandler, handler_args2)

            for route, handler, args in action.extra_handlers():
                assert route != '/' and route.startswith('/')
                handler_args3 = handler_args2.copy()
                handler_args3.update(args)
                server.add_route(f'/actions/{name}' + route, handler, handler_args3)

        server.add_route('/healthz', HealthHandler, handler_args)
        server.add_route(r'/(.*)', Error)

        server.startup(address=config.HOST, port=config.PORT)

        self.server = server

    async def start(self):
        if self.prometheus_port:
            logging.info("starting prometheus on %r", self.prometheus_port)
            start_http_server(self.prometheus_port)
            i = Info('iceprod', 'IceProd information')
            i.info({
                'version': VERSION_STRING,
                'type': 'api',
            })
            self.async_monitor = AsyncMonitor(labels={'type': 'api'})
            await self.async_monitor.start()

        await self.message_queue.setup()

    async def stop(self):
        await self.server.stop()
        if self.async_monitor:
            await self.async_monitor.stop()
        await self.message_queue.close()



"""
Test script for scheduled_tasks/log_cleanup
"""

import logging
from unittest.mock import AsyncMock, MagicMock

import pytest
from iceprod.scheduled_tasks import log_cleanup

logger = logging.getLogger('scheduled_tasks_log_cleanup_test')


async def test_200_run():
    rc = MagicMock()
    logs = {}
    async def client(method, url, args=None):
        logger.info('REST: %s, %s', method, url)
        if method == 'GET' and url.startswith('/logs'):
            return logs
        elif method == 'DELETE' and url.startswith('/logs'):
            client.called = True
        else:
            raise Exception()
    client.called = False
    rc.request = client

    await log_cleanup.run(rc, debug=True)
    assert not client.called

    client.called = False
    logs['a'] = {'log_id':'a'}
    await log_cleanup.run(rc, debug=True)
    assert client.called

async def test_201_run():
    rc = MagicMock()
    rc.request = AsyncMock(side_effect=Exception())
    with pytest.raises(Exception):
        await log_cleanup.run(rc, debug=True)

    # check it normally hides the error
    await log_cleanup.run(rc, debug=False)
